import sys
import os

from typing import Text, Union, Any, Optional, Tuple, List, Dict

if sys.version_info[0] == 2:
    _pathlike = Union[Text, bytes]
else:
    _pathlike = Union[Text, bytes, 'os.PathLike[Any]']
_uri = Union[Text, str]

if sys.version_info[0] == 2:
    if sys.platform == "win32":
        _base = Text
    else:
        _base = bytes
else:
    _base = Text

class fsnative(_base):
    def __init__(self, object: Text=u"") -> None:
        ...

_fsnative = Union[fsnative, _base]

if sys.platform == "win32":
    _bytes_default_encoding = str
else:
    _bytes_default_encoding = Optional[str]

def path2fsn(path: _pathlike) -> _fsnative:
    ...

def fsn2text(path: _fsnative, strict: bool=False) -> Text:
    ...

def text2fsn(text: Text) -> _fsnative:
    ...

def fsn2bytes(path: _fsnative, encoding: _bytes_default_encoding="utf-8") -> bytes:
    ...

def bytes2fsn(data: bytes, encoding: _bytes_default_encoding="utf-8") -> _fsnative:
    ...

def uri2fsn(uri: _uri) -> _fsnative:
    ...

def fsn2uri(path: _fsnative) -> Text:
    ...

def fsn2norm(path: _fsnative) -> _fsnative:
    ...

sep: _fsnative
pathsep: _fsnative
curdir: _fsnative
pardir: _fsnative
altsep: _fsnative
extsep: _fsnative
devnull: _fsnative
defpath: _fsnative

def getcwd() -> _fsnative:
    ...

def getenv(key: _pathlike, value: Optional[_fsnative]=None) -> Optional[_fsnative]:
    ...

def putenv(key: _pathlike, value: _pathlike):
    ...

def unsetenv(key: _pathlike) -> None:
    ...

def supports_ansi_escape_codes(fd: int) -> bool:
    ...

def expandvars(path: _pathlike) -> _fsnative:
    ...

def expanduser(path: _pathlike) -> _fsnative:
    ...

environ: Dict[_fsnative,_fsnative]
argv: List[_fsnative]

def gettempdir() -> _fsnative:
    pass

def mkstemp(suffix: Optional[_pathlike]=None, prefix: Optional[_pathlike]=None, dir: Optional[_pathlike]=None, text: bool=False) -> Tuple[int, _fsnative]:
    ...

def mkdtemp(suffix: Optional[_pathlike]=None, prefix: Optional[_pathlike]=None, dir: Optional[_pathlike]=None) -> _fsnative:
    ...

version_string: str

version: Tuple[int, int, int]

print_ = print

def input_(prompt: Any=None) -> _fsnative:
    ...



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from .misc import get_ca_file

from urllib import request as request_module
from http.client import HTTPException


Request = request_module.Request

UrllibError = EnvironmentError


def urlopen(*args, **kwargs):
    try:
        return request_module.urlopen(*args, **kwargs)
    except HTTPException as e:
        # https://bugs.python.org/issue8823
        raise OSError(e) from e


def install_urllib2_ca_file():
    """Makes urllib2.urlopen and urllib2.build_opener use the ca file
    returned by get_ca_file()
    """

    try:
        import ssl
    except ImportError:
        return

    base = request_module.HTTPSHandler

    class MyHandler(base):
        def __init__(self, debuglevel=0, context=None):
            ca_file = get_ca_file()
            if context is None and ca_file is not None:
                context = ssl.create_default_context(
                    purpose=ssl.Purpose.SERVER_AUTH, cafile=ca_file
                )
            base.__init__(self, debuglevel, context)

    request_module.HTTPSHandler = MyHandler



# Quod Libet Telepathy Plugin
# Copyright 2012 Christoph Reiter
#      2012,2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk

from quodlibet import _
from quodlibet.pattern import Pattern
from quodlibet.qltk.entry import UndoEntry
from quodlibet import util
from quodlibet import qltk

from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins import PluginConfigMixin
from quodlibet.util.dprint import print_d
from quodlibet.qltk import Icons


AM_PATH = "/org/freedesktop/Telepathy/AccountManager"
AM_NAME = "org.freedesktop.Telepathy.AccountManager"
AM_IFACE = "org.freedesktop.Telepathy.AccountManager"
AC_IFACE = "org.freedesktop.Telepathy.Account"
PROPS_IFACE = "org.freedesktop.DBus.Properties"
CONN_PRESENCE_TYPE_AVAILABLE = 2


def is_valid_presence_type(x):
    return x not in [0, 7, 8]


def get_active_account_paths():
    bus_iface = Gio.DBusProxy.new_for_bus_sync(
        Gio.BusType.SESSION,
        Gio.DBusProxyFlags.NONE,
        None,
        AM_NAME,
        AM_PATH,
        PROPS_IFACE,
        None,
    )
    return bus_iface.Get("(ss)", AM_IFACE, "ValidAccounts")


def set_accounts_requested_presence(paths, message):
    bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
    for path in paths:
        bus_iface = Gio.DBusProxy.new_sync(
            bus, Gio.DBusProxyFlags.NONE, None, AM_NAME, path, PROPS_IFACE, None
        )
        presence_type, status = bus_iface.Get("(ss)", AC_IFACE, "CurrentPresence")[:2]
        if not is_valid_presence_type(presence_type):
            presence_type = CONN_PRESENCE_TYPE_AVAILABLE
        value = GLib.Variant("(uss)", (presence_type, status, message))
        bus_iface.Set("(ssv)", AC_IFACE, "RequestedPresence", value)


class TelepathyStatusPlugin(EventPlugin, PluginConfigMixin):
    PLUGIN_ID = "Telepathy Status"
    PLUGIN_NAME = _("Telepathy Status Messages")
    PLUGIN_DESC = _(
        "Updates all Telepathy-based IM accounts (as configured "
        "in Empathy etc) with a status message based on current "
        "song."
    )
    PLUGIN_ICON = Icons.FACE_SMILE

    DEFAULT_PAT = "♫ <~artist~title> ♫"
    DEFAULT_PAT_PAUSED = "<~artist~title> [{}]".format(_("paused"))
    CFG_STATUS_SONGLESS = "no_song_text"
    CFG_LEAVE_STATUS = "leave_status"
    CFG_PAT_PLAYING = "playing_pattern"
    CFG_PAT_PAUSED = "paused_pattern"

    def _set_status(self, text):
        print_d(f'Setting status to "{text}"...')
        self.status = text
        try:
            accounts = get_active_account_paths()
            # TODO: account filtering
            set_accounts_requested_presence(accounts, text)
        except GLib.Error:
            print_d("...but setting failed")
            util.print_exc()

    def plugin_on_song_started(self, song):
        self.song = song
        pat_str = self.config_get(self.CFG_PAT_PLAYING, self.DEFAULT_PAT)
        pattern = Pattern(pat_str)
        status = (
            pattern.format(song)
            if song
            else self.config_get(self.CFG_STATUS_SONGLESS, "")
        )
        self._set_status(status)

    def plugin_on_paused(self):
        pat_str = self.config_get(self.CFG_PAT_PAUSED, self.DEFAULT_PAT_PAUSED)
        pattern = Pattern(pat_str)
        self.status = pattern.format(self.song) if self.song else ""
        self._set_status(self.status)

    def plugin_on_unpaused(self):
        self.plugin_on_song_started(self.song)

    def disabled(self):
        if self.status:
            self._set_status(self.config_get(self.CFG_STATUS_SONGLESS))

    def enabled(self):
        self.song = None
        self.status = ""

    def PluginPreferences(self, parent):
        outer_vb = Gtk.VBox(spacing=12)
        vb = Gtk.VBox(spacing=12)

        # Playing
        hb = Gtk.HBox(spacing=6)
        entry = UndoEntry()
        entry.set_text(self.config_get(self.CFG_PAT_PLAYING, self.DEFAULT_PAT))
        entry.connect("changed", self.config_entry_changed, self.CFG_PAT_PLAYING)
        lbl = Gtk.Label(label=_("Playing:"))
        entry.set_tooltip_markup(
            _("Status text when a song is started. Accepts QL Patterns e.g. %s")
            % util.monospace("<~artist~title>")
        )
        lbl.set_mnemonic_widget(entry)
        hb.pack_start(lbl, False, True, 0)
        hb.pack_start(entry, True, True, 0)
        vb.pack_start(hb, True, True, 0)

        # Paused
        hb = Gtk.HBox(spacing=6)
        entry = UndoEntry()
        entry.set_text(self.config_get(self.CFG_PAT_PAUSED, self.DEFAULT_PAT_PAUSED))
        entry.connect("changed", self.config_entry_changed, self.CFG_PAT_PAUSED)
        lbl = Gtk.Label(label=_("Paused:"))
        entry.set_tooltip_markup(
            _("Status text when a song is paused. Accepts QL Patterns e.g. %s")
            % util.monospace("<~artist~title>")
        )
        lbl.set_mnemonic_widget(entry)
        hb.pack_start(lbl, False, True, 0)
        hb.pack_start(entry, True, True, 0)
        vb.pack_start(hb, True, True, 0)

        # No Song
        hb = Gtk.HBox(spacing=6)
        entry = UndoEntry()
        entry.set_text(self.config_get(self.CFG_STATUS_SONGLESS, ""))
        entry.connect("changed", self.config_entry_changed, self.CFG_STATUS_SONGLESS)
        entry.set_tooltip_text(_("Plain text for status when there is no current song"))
        lbl = Gtk.Label(label=_("No song:"))
        lbl.set_mnemonic_widget(entry)
        hb.pack_start(lbl, False, True, 0)
        hb.pack_start(entry, True, True, 0)
        vb.pack_start(hb, True, True, 0)

        # Frame
        frame = qltk.Frame(_("Status Patterns"), child=vb)
        outer_vb.pack_start(frame, False, True, 0)

        return outer_vb



# Copyright 2012 - 2024 Christoph Reiter, Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from typing import Optional
from collections.abc import Iterable

from quodlibet import _
from quodlibet import config
from quodlibet import util
from quodlibet.qltk.ccb import ConfigCheckButton
from quodlibet.util import escape
from quodlibet.util.config import ConfigProxy
from quodlibet.util.dprint import print_d
from quodlibet.util.modulescanner import ModuleScanner


def init(folders=None, disable_plugins=False):
    """folders: list of paths to look for plugins
    disable_plugins: disables all plugins, but does not forget which
    plugins are enabled.
    """
    if disable_plugins:
        folders = []
    manager = PluginManager.instance = PluginManager(folders)
    return manager


def quit():
    PluginManager.instance.save()
    PluginManager.instance.quit()
    PluginManager.instance = None


class PluginImportError(Exception):
    desc = ""

    def __init__(self, desc, *args, **kwargs):
        super().__init__(desc)
        self.desc = desc

    def should_show(self):
        """If the error description should be shown to the user"""

        return True


class PluginNotSupportedError(PluginImportError):
    """To hide the plugin (e.g. on Windows)"""

    def __init__(self, msg=None):
        msg = "not supported: %s" % (msg or "unknown reason")
        super().__init__(msg)

    def should_show(self):
        return False


class MissingModulePluginError(PluginImportError):
    """Consistent Exception for reporting missing modules for plugins"""

    def __init__(self, module_name: str, extra: str | None = None):
        tmpl = _(
            "Couldn't find module '{module}'. "
            "Perhaps you need to install the package?{extra}"
        )
        msg = tmpl.format(
            module=module_name, extra="" if extra is None else " " + extra
        )
        super().__init__(msg)


class MissingGstreamerElementPluginError(PluginImportError):
    """Consistent Exception for reporting missing Gstreamer elements for
    plugins"""

    def __init__(self, element_name):
        msg = _("Couldn't find GStreamer element '{element}'.").format(
            element=element_name
        )
        super().__init__(msg)


def migrate_old_config():
    active = []
    old_keys = [
        "songsmenuplugins",
        "eventplugins",
        "editingplugins",
        "playorderplugins",
    ]
    for key in old_keys:
        key = "active_" + key
        try:
            active.extend(config.get("plugins", key).splitlines())
        except config.Error:
            pass
        else:
            config._config.remove_option("plugins", key)

    if active:
        config.set("plugins", "active_plugins", "\n".join(sorted(active)))


def list_plugins(module):
    """Return all objects of the module that satisfy the basic
    plugin needs: id, name and don't start with '_'

    If '__all__' is defined, only plugins in '__all__' will be loaded.
    """

    try:
        objs = [getattr(module, attr) for attr in module.__all__]
    except AttributeError:
        objs = [
            getattr(module, attr) for attr in vars(module) if not attr.startswith("_")
        ]

    ok = []
    for obj in objs:
        if hasattr(obj, "PLUGIN_ID"):
            if not hasattr(obj, "PLUGIN_NAME"):
                obj.PLUGIN_NAME = obj.PLUGIN_ID
            ok.append(obj)

    return ok


class PluginModule:
    def __init__(self, name, module):
        self.name = name
        self.module = module
        self.plugins = [Plugin(cls) for cls in list_plugins(module)]


class Plugin:
    def __init__(self, plugin_cls):
        self.cls = plugin_cls
        self.handlers = []
        self.instance = None

    def __repr__(self):
        return f"<{type(self).__name__} id={self.id!r} name={self.name!r}>"

    @property
    def can_enable(self):
        return getattr(self.cls, "PLUGIN_CAN_ENABLE", True)

    @property
    def id(self):
        return self.cls.PLUGIN_ID

    @property
    def name(self):
        return self.cls.PLUGIN_NAME

    @property
    def description(self):
        return getattr(self.cls, "PLUGIN_DESC", None)

    @property
    def description_markup(self):
        try:
            return self.cls.PLUGIN_DESC_MARKUP
        except AttributeError:
            return escape(self.description)

    @property
    def tags(self):
        tags = getattr(self.cls, "PLUGIN_TAGS", [])
        if isinstance(tags, str):
            tags = [tags]
        return tags

    @property
    def icon(self):
        return getattr(self.cls, "PLUGIN_ICON", None)

    def get_instance(self):
        """A singleton"""

        if not getattr(self.cls, "PLUGIN_INSTANCE", False):
            return None

        if self.instance is None:
            try:
                obj = self.cls()
            except Exception:
                util.print_exc()
                return None
            self.instance = obj

        return self.instance


class PluginHandler:
    """A plugin handler can choose to handle plugins, as well as control
    their enabled state."""

    def plugin_handle(self, plugin):
        """Returns `True` IFF this handler can handle `plugin`"""
        raise NotImplementedError

    def plugin_enable(self, plugin):
        """Called to enable / register `plugin`"""
        raise NotImplementedError

    def plugin_disable(self, plugin):
        """Called to disable / de-register `plugin`"""
        raise NotImplementedError


class PluginManager:
    """
    The manager takes care of plugin loading/reloading. Interested plugin
    handlers can register them self to get called when plugins get enabled
    or disabled.

    Plugins get exposed when at least one handler shows interest
    in them (by returning True in the handle method).

    Plugins have to be a class which defines PLUGIN_ID, PLUGIN_NAME.
    Plugins that have a true PLUGIN_INSTANCE attribute get instantiated on
    enable and the enabled/disabled methods get called.

    If plugin handlers want a plugin instance, they have to call
    Plugin.get_instance() to get a singleton.

    handlers need to implement the following methods:

        handler.plugin_handle(plugin)
            Needs to return True if the handler should be called
            whenever the plugin's enabled status changes.

        handler.plugin_enable(plugin)
            Gets called if the plugin gets enabled.

        handler.plugin_disable(plugin)
            Gets called if the plugin gets disabled.
            Should remove all references.
    """

    CONFIG_SECTION = "plugins"
    CONFIG_OPTION = "active_plugins"

    instance: Optional["PluginManager"] = None
    """Default instance"""

    def __init__(self, folders=None):
        """folders is a list of paths that will be scanned for plugins.
        Plugins in later paths will be preferred if they share a name.
        """

        super().__init__()

        if folders is None:
            folders = []

        self.__scanner = ModuleScanner(folders)
        self.__modules = {}  # name: PluginModule
        self.__handlers = []  # handler list
        self.__enabled = set()  # (possibly) enabled plugin IDs

        self.__restore()

    def rescan(self):
        """Scan for plugin changes or to initially load all plugins"""

        print_d("Rescanning..")

        removed, added = self.__scanner.rescan()

        # remember IDs of enabled plugin that get reloaded, so we can enable
        # them again
        reload_ids = []
        for name in removed:
            if name not in added:
                continue
            mod = self.__modules[name]
            for plugin in mod.plugins:
                if self.enabled(plugin):
                    reload_ids.append(plugin.id)

        for name in removed:
            # share the namespace with ModuleScanner for now
            self.__remove_module(name)

        # restore enabled state
        self.__enabled.update(reload_ids)

        for name in added:
            new_module = self.__scanner.modules[name]
            self.__add_module(name, new_module.module)

        print_d("Rescanning done.")

    @property
    def _modules(self):
        return self.__scanner.modules.values()

    @property
    def _plugins(self) -> Iterable[Plugin]:
        """All registered plugins"""
        return (
            plugin for module in self.__modules.values() for plugin in module.plugins
        )

    @property
    def plugins(self):
        """Returns a list of plugins with active handlers"""

        return [p for p in self._plugins if p.handlers]

    def register_handler(self, handler):
        """
        Registers a handler, attaching it to any current plugins it
        advertises that it can handle

        `handler` should probably be a `PluginHandler`
        """
        print_d(f"Registering handler: {type(handler).__name__!r}")

        self.__handlers.append(handler)

        for plugin in self._plugins:
            if not handler.plugin_handle(plugin):
                continue
            if plugin.handlers:
                plugin.handlers.append(handler)
                if self.enabled(plugin):
                    handler.plugin_enable(plugin)
            else:
                plugin.handlers.append(handler)
                if self.enabled(plugin):
                    self.enable(plugin, True, force=True)

    def save(self):
        print_d("Saving plugins: %d active" % len(self.__enabled))
        config.set(
            self.CONFIG_SECTION, self.CONFIG_OPTION, "\n".join(sorted(self.__enabled))
        )

    def enabled(self, plugin):
        """Returns if the plugin is enabled."""

        if not plugin.handlers:
            return False

        return plugin.id in self.__enabled

    def enable(self, plugin, status, force=False):
        """Enable or disable a plugin."""

        if not force and self.enabled(plugin) == bool(status):
            return

        if not status:
            print_d(f"Disable {plugin.id!r}")
            for handler in plugin.handlers:
                handler.plugin_disable(plugin)

            self.__enabled.discard(plugin.id)

            instance = plugin.instance
            if instance and hasattr(instance, "disabled"):
                try:
                    instance.disabled()
                except Exception:
                    util.print_exc()
        else:
            print_d(f"Enable {plugin.id!r}")
            obj = plugin.get_instance()
            if obj and hasattr(obj, "enabled"):
                try:
                    obj.enabled()
                except Exception:
                    util.print_exc()
            for handler in plugin.handlers:
                handler.plugin_enable(plugin)
            self.__enabled.add(plugin.id)

    @property
    def failures(self):
        """module name: list of error message text lines"""

        errors = {}
        for name, error in self.__scanner.failures.items():
            exception = error.exception
            if isinstance(exception, PluginImportError):
                if not exception.should_show():
                    continue
                errors[name] = [exception.desc]
            else:
                errors[name] = error.traceback

        return errors

    def quit(self):
        """Disable plugins and tell all handlers to clean up"""

        for name in list(self.__modules.keys()):
            self.__remove_module(name)

    def __remove_module(self, name):
        plugin_module = self.__modules.pop(name)
        for plugin in plugin_module.plugins:
            if plugin.handlers:
                self.enable(plugin, False)

    def __add_module(self, name, module):
        plugin_mod = PluginModule(name, module)
        self.__modules[name] = plugin_mod

        for plugin in plugin_mod.plugins:
            handlers = []
            for handler in self.__handlers:
                if handler.plugin_handle(plugin):
                    handlers.append(handler)
            if handlers:
                plugin.handlers = handlers
                if self.enabled(plugin):
                    self.enable(plugin, True, force=True)

    def __restore(self):
        migrate_old_config()
        active = config.get(self.CONFIG_SECTION, self.CONFIG_OPTION).splitlines()

        self.__enabled.update(active)
        print_d(f"Restoring {len(self.__enabled)} plugin(s)")

        for plugin in self._plugins:
            if self.enabled(plugin):
                self.enable(plugin, True, force=True)


PM = PluginManager


def plugin_enabled(plugin):
    """Returns true if the plugin is enabled (or "always" enabled)"""
    pm = PluginManager.instance
    return pm.enabled(plugin) or not plugin.can_enable


class PluginConfig(ConfigProxy):
    """A proxy for a Config object that can be used by plugins.

    Provides some methods of the Config class but doesn't need a
    section and prefixes the config option name.
    """

    def __init__(self, prefix, _config=None, _defaults=True):
        self._prefix = prefix
        if _config is None:
            _config = config._config
        super().__init__(_config, PM.CONFIG_SECTION, _defaults)

    def _new_defaults(self, real_default_config):
        return PluginConfig(self._prefix, real_default_config, False)

    def _option(self, name):
        return f"{self._prefix}_{name}"

    def ConfigCheckButton(self, label, option, **kwargs):  # noqa
        return ConfigCheckButton(
            label, PM.CONFIG_SECTION, self._option(option), **kwargs
        )


class PluginConfigMixin:
    """
    Mixin for storage and editing of plugin config in a standard way.
    """

    CONFIG_SECTION = ""
    """If defined, the section for storing config,
        otherwise, it will based on a munged `PLUGIN_ID`"""

    @classmethod
    def _config_key(cls, name):
        return cls._get_config_option(name)

    @classmethod
    def _get_config_option(cls, option):
        prefix = cls.CONFIG_SECTION
        if not prefix:
            prefix = cls.PLUGIN_ID.lower().replace(" ", "_")

        return f"{prefix}_{option}"

    @classmethod
    def config_get(cls, name, default=""):
        """Gets a config string value for this plugin"""
        return config.get(PM.CONFIG_SECTION, cls._config_key(name), default)

    @classmethod
    def config_set(cls, name, value):
        """Saves a config string value for this plugin"""
        try:
            config.set(PM.CONFIG_SECTION, cls._config_key(name), value)
        except config.Error:
            print_d(f"Couldn't set config item '{name}' to {value!r}")

    @classmethod
    def config_get_bool(cls, name, default=False):
        """Gets a config boolean for this plugin"""
        return config.getboolean(PM.CONFIG_SECTION, cls._config_key(name), default)

    @classmethod
    def config_get_stringlist(cls, name, default=False):
        """Gets a config string list for this plugin"""
        return config.getstringlist(PM.CONFIG_SECTION, cls._config_key(name), default)

    def config_entry_changed(self, entry, key):
        """React to a change in a gtk.Entry (by saving it to config)"""
        if entry.get_property("sensitive"):
            self.config_set(key, entry.get_text())

    @classmethod
    def ConfigCheckButton(cls, label, name, default=False):  # noqa
        """
        Create a new `ConfigCheckButton` for `name`, pre-populated correctly
        """
        option = cls._config_key(name)
        try:
            config.getboolean(PM.CONFIG_SECTION, option)
        except config.Error:
            cls.config_set(name, default)
        return ConfigCheckButton(label, PM.CONFIG_SECTION, option, populate=True)


class ConfProp:
    def __init__(self, conf, name, default):
        self._conf = conf
        self._name = name

        self._conf.defaults.set(name, default)

    def __get__(self, *args, **kwargs):
        return self._conf.get(self._name)

    def __set__(self, obj, value):
        self._conf.set(self._name, value)


class BoolConfProp(ConfProp):
    def __get__(self, *args, **kwargs):
        return self._conf.getboolean(self._name)


class IntConfProp(ConfProp):
    def __get__(self, *args, **kwargs):
        return self._conf.getint(self._name)


class FloatConfProp(ConfProp):
    def __get__(self, *args, **kwargs):
        return self._conf.getfloat(self._name)


def str_to_color_tuple(s):
    """Raises ValueError"""

    lst = [float(p) for p in s.split()]
    while len(lst) < 4:
        lst.append(0.0)
    return tuple(lst)


def color_tuple_to_str(t):
    return " ".join(map(str, t))


class ColorConfProp(ConfProp):
    def __init__(self, conf, name, default):
        self._conf = conf
        self._name = name

        self._conf.defaults.set(name, color_tuple_to_str(default))

    def __get__(self, *args, **kwargs):
        s = self._conf.get(self._name)

        try:
            return str_to_color_tuple(s)
        except ValueError:
            return str_to_color_tuple(self._conf.defaults.get(self._name))

    def __set__(self, obj, value):
        self._conf.set(self._name, color_tuple_to_str(value))



# Copyright 2010,2012 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time
import tempfile

import dbus
import dbus.service
from senf import fsn2uri

from quodlibet import app
from quodlibet.order.repeat import RepeatListForever, RepeatSongForever
from quodlibet.qltk.songlist import get_columns
from quodlibet.util.dbusutils import DBusIntrospectable, DBusProperty
from quodlibet.util.dbusutils import dbus_unicode_validate as unival

from .util import MPRISObject

# TODO: OpenUri, CanXYZ
# Date parsing (util?)


# http://www.mpris.org/2.0/spec/
class MPRIS2(DBusProperty, DBusIntrospectable, MPRISObject):
    BUS_NAME = "org.mpris.MediaPlayer2.quodlibet"
    PATH = "/org/mpris/MediaPlayer2"

    ROOT_IFACE = "org.mpris.MediaPlayer2"

    ROOT_ISPEC = """
<method name="Raise"/>
<method name="Quit"/>"""

    ROOT_PROPS = """
<property name="CanQuit" type="b" access="read"/>
<property name="CanRaise" type="b" access="read"/>
<property name="CanSetFullscreen" type="b" access="read"/>
<property name="HasTrackList" type="b" access="read"/>
<property name="Identity" type="s" access="read"/>
<property name="DesktopEntry" type="s" access="read"/>
<property name="SupportedUriSchemes" type="as" access="read"/>
<property name="SupportedMimeTypes" type="as" access="read"/>"""

    PLAYER_IFACE = "org.mpris.MediaPlayer2.Player"

    PLAYER_ISPEC = """
<method name="Next"/>
<method name="Previous"/>
<method name="Pause"/>
<method name="PlayPause"/>
<method name="Stop"/>
<method name="Play"/>
<method name="Seek">
  <arg direction="in" name="Offset" type="x"/>
</method>
<method name="SetPosition">
  <arg direction="in" name="TrackId" type="o"/>
  <arg direction="in" name="Position" type="x"/>
</method>
<method name="OpenUri">
  <arg direction="in" name="Uri" type="s"/>
</method>
<signal name="Seeked">
  <arg name="Position" type="x"/>
</signal>"""

    PLAYER_PROPS = """
<property name="PlaybackStatus" type="s" access="read"/>
<property name="LoopStatus" type="s" access="readwrite"/>
<property name="Rate" type="d" access="readwrite"/>
<property name="Shuffle" type="b" access="readwrite"/>
<property name="Metadata" type="a{sv}" access="read"/>
<property name="Volume" type="d" access="readwrite"/>
<property name="Position" type="x" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" \
value="false"/>
</property>
<property name="MinimumRate" type="d" access="read"/>
<property name="MaximumRate" type="d" access="read"/>
<property name="CanGoNext" type="b" access="read"/>
<property name="CanGoPrevious" type="b" access="read"/>
<property name="CanPlay" type="b" access="read"/>
<property name="CanPause" type="b" access="read"/>
<property name="CanSeek" type="b" access="read"/>
<property name="CanControl" type="b" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" \
value="false"/>
</property>"""

    def __init__(self):
        DBusIntrospectable.__init__(self)
        DBusProperty.__init__(self)

        self.set_introspection(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_ISPEC)
        self.set_properties(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_PROPS)

        self.set_introspection(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_ISPEC)
        self.set_properties(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_PROPS)

        bus = dbus.SessionBus()
        name = dbus.service.BusName(self.BUS_NAME, bus)
        MPRISObject.__init__(self, bus, self.PATH, name)

        self.__metadata = None
        self.__cover = None
        player_options = app.player_options
        self.__repeat_id = player_options.connect(
            "notify::repeat", self.__repeat_changed
        )
        self.__random_id = player_options.connect(
            "notify::shuffle", self.__shuffle_changed
        )
        self.__single_id = player_options.connect(
            "notify::single", self.__single_changed
        )

        self.__lsig = app.librarian.connect("changed", self.__library_changed)
        self.__vsig = app.player.connect("notify::volume", self.__volume_changed)
        self.__seek_sig = app.player.connect("seek", self.__seeked)

    def remove_from_connection(self, *arg, **kwargs):
        super().remove_from_connection(*arg, **kwargs)

        player_options = app.player_options
        player_options.disconnect(self.__repeat_id)
        player_options.disconnect(self.__random_id)
        player_options.disconnect(self.__single_id)
        app.librarian.disconnect(self.__lsig)
        app.player.disconnect(self.__vsig)
        app.player.disconnect(self.__seek_sig)

        if self.__cover is not None:
            self.__cover.close()
            self.__cover = None
        self.__invalidate_metadata()

    def __volume_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Volume"])

    def __repeat_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["LoopStatus"])

    def __shuffle_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Shuffle"])

    def __single_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["LoopStatus"])

    def __seeked(self, player, song, ms):
        self.Seeked(ms * 1000)

    def __library_changed(self, library, songs):
        self.__invalidate_metadata()
        if not songs or app.player.info not in songs:
            return
        self.emit_properties_changed(self.PLAYER_IFACE, ["Metadata"])

    @dbus.service.method(ROOT_IFACE)
    def Raise(self):
        app.present()

    @dbus.service.method(ROOT_IFACE)
    def Quit(self):
        app.quit()

    @dbus.service.signal(PLAYER_IFACE, signature="x")
    def Seeked(self, position):
        pass

    @dbus.service.method(PLAYER_IFACE)
    def Next(self):
        player = app.player
        paused = player.paused
        player.next()
        player.paused = paused

    @dbus.service.method(PLAYER_IFACE)
    def Previous(self):
        player = app.player
        paused = player.paused
        player.previous()
        player.paused = paused

    @dbus.service.method(PLAYER_IFACE)
    def Pause(self):
        app.player.paused = True

    @dbus.service.method(PLAYER_IFACE)
    def Play(self):
        app.player.play()

    @dbus.service.method(PLAYER_IFACE)
    def PlayPause(self):
        app.player.playpause()

    @dbus.service.method(PLAYER_IFACE)
    def Stop(self):
        app.player.stop()

    @dbus.service.method(PLAYER_IFACE, in_signature="x")
    def Seek(self, offset):
        new_pos = app.player.get_position() + offset / 1000
        app.player.seek(new_pos)

    @dbus.service.method(PLAYER_IFACE, in_signature="ox")
    def SetPosition(self, track_id, position):
        if track_id == self.__get_current_track_id():
            app.player.seek(position / 1000)

    def paused(self):
        self.emit_properties_changed(self.PLAYER_IFACE, ["PlaybackStatus"])

    unpaused = paused

    def song_started(self, song):
        self.__invalidate_metadata()

        # so the position in clients gets updated faster
        self.Seeked(0)

        self.emit_properties_changed(self.PLAYER_IFACE, ["PlaybackStatus", "Metadata"])

    def __get_current_track_id(self):
        path = "/net/sacredchao/QuodLibet"
        if not app.player.info:
            return dbus.ObjectPath(path + "/" + "NoTrack")
        return dbus.ObjectPath(path + "/" + str(id(app.player.info)))

    def __invalidate_metadata(self):
        self.__metadata = None

    def __get_metadata(self):
        if self.__metadata is None:
            self.__metadata = self.__get_metadata_real()
            assert self.__metadata is not None
        return self.__metadata

    def __get_metadata_real(self):
        """
        https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/
        """

        metadata = {}
        metadata["mpris:trackid"] = self.__get_current_track_id()

        def ignore_overflow(dbus_type, value):
            try:
                return dbus_type(value)
            except OverflowError:
                return 0

        song = app.player.info
        if not song:
            return metadata

        metadata["mpris:length"] = ignore_overflow(dbus.Int64, song("~#length") * 10**6)

        if self.__cover is not None:
            self.__cover.close()
            self.__cover = None

        cover = app.cover_manager.get_cover(song)
        if cover:
            is_temp = cover.name.startswith(tempfile.gettempdir())
            if is_temp:
                self.__cover = cover
            metadata["mpris:artUrl"] = fsn2uri(cover.name)

        # All list values
        list_val = {
            "artist": "artist",
            "albumArtist": "albumartist",
            "comment": "comment",
            "composer": "composer",
            "genre": "genre",
            "lyricist": "lyricist",
        }
        for xesam, tag in list_val.items():
            vals = song.list(tag)
            if vals:
                metadata["xesam:" + xesam] = list(map(unival, vals))

        # All single values
        columns = get_columns()
        title_tag = "~title~version" if "~title~version" in columns else "title"
        album_tag = (
            "~album~discsubtitle" if "~album~discsubtitle" in columns else "album"
        )

        sing_val = {"album": album_tag, "title": title_tag, "asText": "~lyrics"}
        for xesam, tag in sing_val.items():
            vals = song.comma(tag)
            if vals:
                metadata["xesam:" + xesam] = unival(vals)

        # URI
        metadata["xesam:url"] = song("~uri")

        # Integers
        num_val = {
            "audioBPM": "bpm",
            "discNumber": "disc",
            "trackNumber": "track",
            "useCount": "playcount",
        }

        for xesam, tag in num_val.items():
            val = song("~#" + tag, None)
            if val is not None:
                metadata["xesam:" + xesam] = ignore_overflow(dbus.Int32, val)

        # Rating
        metadata["xesam:userRating"] = ignore_overflow(dbus.Double, song("~#rating"))

        # Dates
        iso_8601_format = "%Y-%m-%dT%H:%M:%S"
        tuple_time = time.gmtime(song("~#lastplayed"))
        iso_time = time.strftime(iso_8601_format, tuple_time)
        metadata["xesam:lastUsed"] = iso_time

        year = song("~year")
        if year:
            try:
                tuple_time = time.strptime(year, "%Y")
                iso_time = time.strftime(iso_8601_format, tuple_time)
            except ValueError:
                pass
            else:
                metadata["xesam:contentCreated"] = iso_time

        return metadata

    def set_property(self, interface, name, value):
        player = app.player
        player_options = app.player_options

        if interface == self.PLAYER_IFACE:
            if name == "LoopStatus":
                if value == "Playlist":
                    player_options.repeat = True
                    app.window.order.repeater = RepeatListForever
                elif value == "Track":
                    player_options.repeat = True
                    app.window.order.repeater = RepeatSongForever
                elif value == "None":
                    player_options.repeat = False
                    app.window.order.repeater = RepeatListForever
            elif name == "Rate":
                pass
            elif name == "Shuffle":
                player_options.shuffle = value
            elif name == "Volume":
                player.volume = value

    def get_property(self, interface, name):
        player = app.player
        player_options = app.player_options

        if interface == self.ROOT_IFACE:
            if name == "CanQuit":
                return True
            if name == "CanRaise":
                return True
            if name == "CanSetFullscreen":
                return False
            if name == "HasTrackList":
                return False
            if name == "Identity":
                return app.name
            if name == "DesktopEntry":
                return "io.github.quodlibet.QuodLibet"
            if name == "SupportedUriSchemes":
                # TODO: enable once OpenUri is done
                def can(s):
                    return False

                schemes = ["http", "https", "ftp", "file", "mms"]
                return filter(can, schemes)
            if name == "SupportedMimeTypes":
                from quodlibet import formats

                return formats.mimes
        elif interface == self.PLAYER_IFACE:
            if name == "PlaybackStatus":
                if not player.song:
                    return "Stopped"
                return ("Playing", "Paused")[int(player.paused)]
            if name == "LoopStatus":
                if not player_options.repeat:
                    return "None"
                if player_options.single:
                    return "Track"
                return "Playlist"
            if name == "Rate":
                return 1.0
            if name == "Shuffle":
                return player_options.shuffle
            if name == "Metadata":
                return self.__get_metadata()
            if name == "Volume":
                # https://gitlab.freedesktop.org/mpris/mpris-spec/issues/8
                return player.volume
            if name == "Position":
                return player.get_position() * 1000
            if name == "MinimumRate":
                return 1.0
            if name == "MaximumRate":
                return 1.0
            if name == "CanGoNext":
                return True
            if name == "CanGoPrevious":
                return True
            if name == "CanPlay":
                return True
            if name == "CanPause":
                return True
            if name == "CanSeek":
                return True
            if name == "CanControl":
                return True
        return None



# Copyright 2016-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import GLib

from quodlibet import config, _
from quodlibet.browsers.soundcloud.query import SoundcloudQuery
from quodlibet.formats import AudioFile
from quodlibet.formats.remote import RemoteFile
from quodlibet.library import SongLibrary
from quodlibet.library.base import K
from quodlibet.qltk.notif import Task
from quodlibet.util import cached_property, print_exc, copool
from quodlibet.util.dprint import print_d, print_w
from senf import fsnative


class SoundcloudLibrary(SongLibrary[K, "SoundcloudFile"]):
    STAR = ["artist", "title", "genre", "tags"]

    def __init__(self, client, player=None):
        super().__init__("Soundcloud")
        self.client = client
        self._sids = [
            self.client.connect("songs-received", self._on_songs_received),
            self.client.connect("stream-uri-received", self._on_stream_uri_received),
            self.client.connect("comments-received", self._on_comments_received),
        ]
        self._psid = None
        # Keep track of async-changed songs for bulk signalling
        self._dirty = set()
        GLib.timeout_add(2000, self._on_tick)
        if player:
            self.player = player
            self._psid = self.player.connect("song-started", self.__song_started)

    def destroy(self):
        super().destroy()
        for sid in self._sids:
            self.client.disconnect(sid)
        if self._psid:
            self.player.disconnect(self._psid)

    def query(self, text, sort=None, star=STAR):
        values = self._contents.values()
        try:
            return SoundcloudQuery(text).filter(values)
        except SoundcloudQuery.Error:
            return values

    def query_with_refresh(self, query: SoundcloudQuery):
        """Queries Soundcloud for some (more) relevant results, then filters"""
        current = self._contents.values()

        if not query.is_parsable:
            return current
        self.client.get_tracks(query.terms)
        filtered = query.filter(current)
        print_d("Filtered %d results to %d" % (len(current), len(filtered)))
        return filtered

    def rename(self, song, newname, changed=None):
        raise TypeError("Can't rename Soundcloud files")

    def _get_stream_urls(self, songs):
        # Pre-cache. It's horrible, but at least you can play things immediately
        with Task(_("Soundcloud"), "Pre-fetching stream URLs") as task:
            total = len(songs)
            for i, song in enumerate(songs):
                # Only update ones without streaming URLs
                # TODO: But yes these will time out...
                if "~uri" not in song or "api.soundcloud.com" in song["~uri"]:
                    self.client.get_stream_url(song)
                task.update(i / total)
                yield

    def _on_songs_received(self, client, songs):
        print_d(f"Got {len(songs)} songs")
        self.add(songs)
        # Can't have multiple requests cancel each other's copools
        funcid = hash("".join(s["~uri"] for s in songs))
        # Rate limit a little to avoid 429s
        copool.add(self._get_stream_urls, songs, timeout=100, funcid=funcid)

    def _on_stream_uri_received(self, client, song: AudioFile, uri: str):
        # URI isn't the key in this SoundcloudFile, so this is OK
        song["~uri"] = uri
        self._dirty.add(song)

    def _on_tick(self) -> bool:
        if self._dirty:
            self.changed(self._dirty)
            self._dirty.clear()
        return True

    def _on_comments_received(self, client, track_id, comments):
        def bookmark_for(com):
            text = f'{com["body"]!r} – {com["user"]["username"]}'
            return max(0, int((com.get("timestamp") or 0) / 1000.0)), text

        try:
            song = self.song_by_track_id(track_id)
        except KeyError:
            # https://github.com/quodlibet/quodlibet/issues/2410
            print_exc()
            return
        song.bookmarks = [bookmark_for(c) for c in comments]
        print_d(f"Updated song bookmarks for {song('title')}")

    def song_by_track_id(self, track_id):
        for song in self.values():
            if song.track_id == track_id:
                return song
        raise KeyError(
            f"No track with id {track_id}. "
            f"Do have {[s.track_id for s in self.values()]}"
        )

    def _changed(self, items):
        super()._changed(items)
        # We should ask the AudioFile subclass to write what it can ASAP
        for item in items:
            item.write()

    def __song_started(self, player, song):
        if isinstance(song, SoundcloudFile):
            print_d(f"Getting comments for {song('title')} ({song.key})")
            self.client.get_comments(song.track_id)


class SoundcloudFile(RemoteFile):
    format = "Remote Soundcloud File"

    def __init__(self, uri: str, track_id: int, client, favorite: bool = False):
        # Don't call super, it invokes __getitem__
        self["~uri"] = uri
        self.sanitize(fsnative(uri))
        self.client = client
        if not self.client:
            raise OSError("Must have a Soundcloud client")
        self["soundcloud_track_id"] = track_id
        self.favorite = favorite
        if self.favorite:
            self["~#rating"] = 1.0

    def set_image(self, image):
        raise TypeError("Can't change images on Soundcloud")

    @cached_property
    def track_id(self):
        return int(self["soundcloud_track_id"])

    @cached_property
    def key(self):
        return f"track-{self.track_id}"

    def can_change(self, k=None):
        if k is None:
            return ["~rating", "~#rating"]
        return k.endswith("rating")

    def write(self):
        if not self.client or not self.client.online:
            print_w("Can't save without a logged-in Soundcloud client")
            return
        # There's not much that can be written
        self._write_rating()

    def _write_rating(self):
        should_fave = self.has_rating and self("~#rating") >= config.RATINGS.default
        track_id = self.track_id
        if not self.favorite and should_fave:
            self.client.save_favorite(track_id)
            self.favorite = True
        elif self.favorite and not should_fave:
            self.client.remove_favorite(track_id)
            self.favorite = False



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""
Error handling: every function which has ctypes.HRESULT as restype
will automatically raise WindowsError if a bad result is returned.
For all other functions check the return status and raise ctypes.WinError()
"""

import sys
import ctypes

from .enum import enum


if sys.platform == "win32":
    from ctypes import wintypes, cdll, windll, oledll

    class GUID(ctypes.Structure):
        # https://msdn.microsoft.com/en-us/library/windows/desktop/
        #   aa373931%28v=vs.85%29.aspx

        _fields_ = [
            ("Data1", wintypes.DWORD),
            ("Data2", wintypes.WORD),
            ("Data3", wintypes.WORD),
            ("Data4", wintypes.BYTE * 8),
        ]

        def __init__(self, name=None):
            if name is not None:
                IIDFromString(str(name), ctypes.byref(self))

        def __str__(self):
            ptr = wintypes.LPOLESTR()
            StringFromIID(ctypes.byref(self), ctypes.byref(ptr))
            string = str(ptr.value)
            CoTaskMemFree(ptr)
            return string

    LPGUID = ctypes.POINTER(GUID)
    IID = GUID
    LPIID = ctypes.POINTER(IID)
    REFIID = ctypes.POINTER(IID)
    CLSID = GUID
    REFCLSID = ctypes.POINTER(CLSID)

    WORD = wintypes.WORD
    DWORD = wintypes.DWORD
    ULONG = wintypes.ULONG
    SHORT = wintypes.SHORT

    ULONG_PTR = wintypes.WPARAM
    LONG_PTR = wintypes.LPARAM
    LRESULT = LONG_PTR
    HHOOK = wintypes.HANDLE
    HRESULT = ctypes.HRESULT

    HC_ACTION = 0
    HC_NOREMOVE = 3

    VK_MEDIA_NEXT_TRACK = 0xB0
    VK_MEDIA_PREV_TRACK = 0xB1
    VK_MEDIA_STOP = 0xB2
    VK_MEDIA_PLAY_PAUSE = 0xB3

    WM_KEYDOWN = 0x0100
    WM_KEYUP = 0x0101
    WM_SYSKEYDOWN = 0x0104
    WM_SYSKEYUP = 0x0105

    CallNextHookEx = windll.user32.CallNextHookEx
    CallNextHookEx.argtypes = [HHOOK, ctypes.c_int, wintypes.WPARAM, wintypes.LPARAM]
    CallNextHookEx.restype = LRESULT

    class KBDLLHOOKSTRUCT(ctypes.Structure):
        _fields_ = [
            ("vkCode", DWORD),
            ("scanCode", DWORD),
            ("flags", DWORD),
            ("time", DWORD),
            ("dwExtraInfo", ULONG_PTR),
        ]

    LPKBDLLHOOKSTRUCT = PKBDLLHOOKSTRUCT = ctypes.POINTER(KBDLLHOOKSTRUCT)

    LowLevelKeyboardProc = ctypes.WINFUNCTYPE(
        LRESULT, ctypes.c_int, wintypes.WPARAM, wintypes.LPARAM
    )

    _SetWindowsHookExW = windll.user32.SetWindowsHookExW
    _SetWindowsHookExW.argtypes = [
        ctypes.c_int,
        ctypes.c_void_p,
        ctypes.c_void_p,
        wintypes.DWORD,
    ]
    _SetWindowsHookExW.restype = HHOOK

    def SetWindowsHookExW(idHook, lpfn, hMod, dwThreadId):
        assert idHook == WH_KEYBOARD_LL
        assert isinstance(lpfn, LowLevelKeyboardProc)
        return _SetWindowsHookExW(idHook, lpfn, hMod, dwThreadId)

    UnhookWindowsHookEx = windll.user32.UnhookWindowsHookEx
    UnhookWindowsHookEx.argtypes = [HHOOK]
    UnhookWindowsHookEx.restype = wintypes.BOOL

    WH_KEYBOARD_LL = 13

    LPWIN32_FIND_DATAW = ctypes.POINTER(wintypes.WIN32_FIND_DATAW)

    IIDFromString = windll.ole32.IIDFromString
    IIDFromString.argtypes = [wintypes.LPCOLESTR, LPIID]
    IIDFromString.restype = HRESULT

    StringFromIID = windll.ole32.StringFromIID
    StringFromIID.argtypes = [REFIID, ctypes.POINTER(wintypes.LPOLESTR)]
    StringFromIID.restype = HRESULT

    CoInitialize = windll.ole32.CoInitialize
    CoInitialize.argtypes = [wintypes.LPVOID]
    CoInitialize.restype = HRESULT

    LPDWORD = ctypes.POINTER(wintypes.DWORD)
    REFKNOWNFOLDERID = ctypes.POINTER(GUID)

    CLSCTX_INPROC_SERVER = 1

    SetEnvironmentVariableW = ctypes.windll.kernel32.SetEnvironmentVariableW
    SetEnvironmentVariableW.argtypes = [ctypes.c_wchar_p, ctypes.c_wchar_p]
    SetEnvironmentVariableW.restype = ctypes.c_bool

    GetEnvironmentStringsW = ctypes.windll.kernel32.GetEnvironmentStringsW
    GetEnvironmentStringsW.argtypes = []
    GetEnvironmentStringsW.restype = ctypes.c_void_p

    FreeEnvironmentStringsW = ctypes.windll.kernel32.FreeEnvironmentStringsW
    FreeEnvironmentStringsW.argtypes = [ctypes.c_void_p]
    FreeEnvironmentStringsW.restype = ctypes.c_bool

    SHGetFolderPathW = ctypes.windll.shell32.SHGetFolderPathW
    SHGetFolderPathW.argtypes = [
        wintypes.HWND,
        ctypes.c_int,
        wintypes.HANDLE,
        wintypes.DWORD,
        wintypes.LPWSTR,
    ]
    SHGetFolderPathW.restype = HRESULT

    SHGetKnownFolderPath = windll.shell32.SHGetKnownFolderPath
    SHGetKnownFolderPath.argtypes = [
        REFKNOWNFOLDERID,
        wintypes.DWORD,
        wintypes.HANDLE,
        ctypes.POINTER(ctypes.c_wchar_p),
    ]
    SHGetKnownFolderPath.restype = HRESULT

    CoTaskMemFree = windll.ole32.CoTaskMemFree
    CoTaskMemFree.argtypes = [ctypes.c_void_p]
    CoTaskMemFree.restype = None

    GetCommandLineW = cdll.kernel32.GetCommandLineW
    GetCommandLineW.argtypes = []
    GetCommandLineW.restype = wintypes.LPCWSTR

    CommandLineToArgvW = windll.shell32.CommandLineToArgvW
    CommandLineToArgvW.argtypes = [wintypes.LPCWSTR, ctypes.POINTER(ctypes.c_int)]
    CommandLineToArgvW.restype = ctypes.POINTER(wintypes.LPWSTR)

    LocalFree = windll.kernel32.LocalFree
    LocalFree.argtypes = [wintypes.HLOCAL]
    LocalFree.restype = wintypes.HLOCAL

    WaitNamedPipeW = windll.kernel32.WaitNamedPipeW
    WaitNamedPipeW.argtypes = [wintypes.LPCWSTR, wintypes.DWORD]
    WaitNamedPipeW.restype = wintypes.BOOL

    LANGID = wintypes.WORD

    GetUserDefaultUILanguage = ctypes.windll.kernel32.GetUserDefaultUILanguage
    GetUserDefaultUILanguage.argtypes = []
    GetUserDefaultUILanguage.restype = LANGID

    GetSystemDefaultUILanguage = ctypes.windll.kernel32.GetSystemDefaultUILanguage
    GetSystemDefaultUILanguage.argtypes = []
    GetSystemDefaultUILanguage.restype = LANGID

    class SECURITY_ATTRIBUTES(ctypes.Structure):
        _fields_ = [
            ("nLength", wintypes.DWORD),
            ("lpSecurityDescriptor", wintypes.LPVOID),
            ("bInheritHandle", wintypes.BOOL),
        ]

    LPSECURITY_ATTRIBUTES = ctypes.POINTER(SECURITY_ATTRIBUTES)
    PSECURITY_ATTRIBUTES = LPSECURITY_ATTRIBUTES

    CreateNamedPipeW = windll.kernel32.CreateNamedPipeW
    CreateNamedPipeW.argtypes = [
        wintypes.LPCWSTR,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.DWORD,
        wintypes.DWORD,
        LPSECURITY_ATTRIBUTES,
    ]
    CreateNamedPipeW.restype = wintypes.HANDLE

    LPOVERLAPPED = ctypes.c_void_p

    PIPE_ACCEPT_REMOTE_CLIENTS = 0x00000000
    PIPE_REJECT_REMOTE_CLIENTS = 0x00000008

    PIPE_ACCESS_DUPLEX = 0x00000003
    PIPE_ACCESS_INBOUND = 0x00000001
    PIPE_ACCESS_OUTBOUND = 0x00000002

    PIPE_TYPE_BYTE = 0x00000000
    PIPE_TYPE_MESSAGE = 0x00000004

    PIPE_READMODE_BYTE = 0x00000000
    PIPE_READMODE_MESSAGE = 0x00000002

    PIPE_WAIT = 0x00000000
    PIPE_NOWAIT = 0x00000001

    FILE_FLAG_FIRST_PIPE_INSTANCE = 0x00080000
    FILE_FLAG_WRITE_THROUGH = 0x80000000
    FILE_FLAG_OVERLAPPED = 0x40000000

    PIPE_UNLIMITED_INSTANCES = 255

    NMPWAIT_USE_DEFAULT_WAIT = 0x00000000
    NMPWAIT_WAIT_FOREVER = 0xFFFFFFFF

    ConnectNamedPipe = windll.kernel32.ConnectNamedPipe
    ConnectNamedPipe.argtypes = [wintypes.HANDLE, LPOVERLAPPED]
    ConnectNamedPipe.restype = wintypes.BOOL

    DisconnectNamedPipe = windll.kernel32.DisconnectNamedPipe
    DisconnectNamedPipe.argtypes = [wintypes.HANDLE]
    DisconnectNamedPipe.restype = wintypes.BOOL

    ReadFile = windll.kernel32.ReadFile
    ReadFile.argtypes = [
        wintypes.HANDLE,
        wintypes.LPVOID,
        wintypes.DWORD,
        LPDWORD,
        LPOVERLAPPED,
    ]
    ReadFile.restype = wintypes.BOOL

    CloseHandle = windll.kernel32.CloseHandle
    CloseHandle.argtypes = [wintypes.HANDLE]
    CloseHandle.restype = wintypes.BOOL

    MOVEFILE_WRITE_THROUGH = 0x8
    MOVEFILE_REPLACE_EXISTING = 0x1

    MoveFileExW = windll.kernel32.MoveFileExW
    MoveFileExW.argtypes = [wintypes.LPWSTR, wintypes.LPWSTR, wintypes.DWORD]
    MoveFileExW.restype = wintypes.BOOL

    GetStdHandle = windll.kernel32.GetStdHandle
    GetStdHandle.argtypes = [DWORD]
    GetStdHandle.restype = wintypes.HANDLE

    SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
    SetConsoleTextAttribute.argtypes = [wintypes.HANDLE, WORD]
    SetConsoleTextAttribute.restype = wintypes.BOOL

    GetConsoleOutputCP = windll.kernel32.GetConsoleOutputCP
    GetConsoleOutputCP.argtypes = []
    GetConsoleOutputCP.restype = wintypes.UINT

    SetConsoleOutputCP = windll.kernel32.SetConsoleOutputCP
    SetConsoleOutputCP.argtypes = [wintypes.UINT]
    SetConsoleOutputCP.restype = wintypes.BOOL

    WinError = ctypes.WinError
    S_OK = HRESULT(0).value
    MAX_PATH = wintypes.MAX_PATH
    INVALID_HANDLE_VALUE = wintypes.HANDLE(-1).value

    STD_INPUT_HANDLE = DWORD(-10)
    STD_OUTPUT_HANDLE = DWORD(-11)
    STD_ERROR_HANDLE = DWORD(-12)

    class COORD(ctypes.Structure):
        _fields_ = [
            ("X", SHORT),
            ("Y", SHORT),
        ]

    class SMALL_RECT(ctypes.Structure):
        _fields_ = [
            ("Left", SHORT),
            ("Top", SHORT),
            ("Right", SHORT),
            ("Bottom", SHORT),
        ]

    class PCONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
        _fields_ = [
            ("dwSize", COORD),
            ("dwCursorPosition", COORD),
            ("wAttributes", WORD),
            ("srWindow", SMALL_RECT),
            ("dwMaximumWindowSize", COORD),
        ]

    GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
    GetConsoleScreenBufferInfo.argtypes = [
        wintypes.HANDLE,
        ctypes.POINTER(PCONSOLE_SCREEN_BUFFER_INFO),
    ]
    GetConsoleScreenBufferInfo.restype = wintypes.BOOL

    FOREGROUND_BLUE = 0x0001
    FOREGROUND_GREEN = 0x0002
    FOREGROUND_RED = 0x0004
    FOREGROUND_INTENSITY = 0x0008
    BACKGROUND_BLUE = 0x0010
    BACKGROUND_GREEN = 0x0020
    BACKGROUND_RED = 0x0040
    BACKGROUND_RED = 0x0040

    class COMMethod:
        def __init__(self, name, offset, restype, argtypes):
            self._name = name
            self._restype = restype
            self._offset = offset
            self._argtypes = argtypes

        def __get__(self, instance, owner):
            func = ctypes.WINFUNCTYPE(self._restype, *self._argtypes)(
                self._offset, self._name
            )

            def wrapper(self, *args, **kwargs):
                return func(self, *args, **kwargs)

            setattr(owner, self._name, wrapper)
            return getattr(instance or owner, self._name)

    class COMInterface(type(ctypes.c_void_p)):
        def __new__(cls, cls_name, bases, d):
            offset = 0
            for base in bases:
                for realbase in base.__mro__:
                    offset += len(realbase.__dict__.get("_methods_", []))

            for i, args in enumerate(d.get("_methods_", [])):
                name = args[0]
                restype = args[1]
                if restype is None:
                    continue
                argtypes = args[2:]
                m = COMMethod(name, offset + i, restype, argtypes)
                d[name] = m

            return type(ctypes.c_void_p).__new__(cls, cls_name, bases, dict(d))

    class IUnknown(ctypes.c_void_p, metaclass=COMInterface):
        IID = GUID("{00000001-0000-0000-c000-000000000046}")

        _methods_ = [
            ("QueryInterface", HRESULT, LPGUID, wintypes.LPVOID),
            ("AddRef", wintypes.DWORD),
            ("Release", wintypes.DWORD),
        ]

    LPUNKNOWN = ctypes.POINTER(IUnknown)

    CoCreateInstance = windll.ole32.CoCreateInstance
    CoCreateInstance.argtypes = [
        REFCLSID,
        LPUNKNOWN,
        wintypes.DWORD,
        REFIID,
        wintypes.LPVOID,
    ]
    CoCreateInstance.restype = HRESULT

    class IShellLinkW(IUnknown):
        IID = GUID("{000214F9-0000-0000-C000-000000000046}")

        _methods_ = [
            (
                "GetPath",
                HRESULT,
                wintypes.LPWSTR,
                wintypes.INT,
                LPWIN32_FIND_DATAW,
                wintypes.DWORD,
            ),
        ]

    class IPersist(IUnknown):
        IID = GUID("{0000010c-0000-0000-C000-000000000046}")

        _methods_ = [
            ("GetClassID", HRESULT, LPGUID),
        ]

    class IPersistFile(IPersist):
        IID = GUID("{0000010b-0000-0000-c000-000000000046}")

        _methods_ = [
            ("IsDirty", HRESULT),
            ("Load", HRESULT, wintypes.LPOLESTR, wintypes.DWORD),
        ]

    class ITEMIDLIST(ctypes.Structure):
        pass

    IBindCtx = ctypes.c_void_p
    ITEMIDLIST_ABSOLUTE = ITEMIDLIST
    ITEMIDLIST_RELATIVE = ITEMIDLIST
    PIDLIST_ABSOLUTE = ctypes.POINTER(ITEMIDLIST_ABSOLUTE)
    PCUIDLIST_RELATIVE = ctypes.POINTER(ITEMIDLIST_RELATIVE)
    PIDLIST_RELATIVE = ctypes.POINTER(ITEMIDLIST_RELATIVE)
    PCIDLIST_ABSOLUTE = ctypes.POINTER(ITEMIDLIST_ABSOLUTE)
    ITEMID_CHILD = ITEMIDLIST
    PCUITEMID_CHILD = ctypes.POINTER(ITEMID_CHILD)
    PCUITEMID_CHILD_ARRAY = ctypes.POINTER(PCUITEMID_CHILD)

    class IShellFolder(IUnknown):
        IID = GUID("{000214E6-0000-0000-C000-000000000046}")

        _methods_ = [
            (
                "ParseDisplayName",
                HRESULT,
                wintypes.HWND,
                ctypes.POINTER(IBindCtx),
                wintypes.LPWSTR,
                ctypes.POINTER(wintypes.ULONG),
                ctypes.POINTER(PIDLIST_RELATIVE),
                ctypes.POINTER(wintypes.ULONG),
            ),
            ("EnumObjects", None),
            (
                "BindToObject",
                HRESULT,
                PCUIDLIST_RELATIVE,
                ctypes.POINTER(IBindCtx),
                REFIID,
                ctypes.c_void_p,
            ),
        ]

    CLSID_ShellLink = GUID("{00021401-0000-0000-C000-000000000046}")

    SHGetDesktopFolder = oledll.shell32.SHGetDesktopFolder
    SHGetDesktopFolder.argtypes = [ctypes.POINTER(IShellFolder)]
    SHGetDesktopFolder.restype = HRESULT

    ILCombine = windll.shell32.ILCombine
    ILCombine.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    ILCombine.restype = ctypes.c_void_p

    ILCreateFromPathW = windll.shell32.ILCreateFromPathW
    ILCreateFromPathW.argtypes = [wintypes.LPCWSTR]
    ILCreateFromPathW.restype = PIDLIST_ABSOLUTE

    ILFree = windll.shell32.ILFree
    ILFree.argtypes = [PIDLIST_RELATIVE]
    ILFree.restype = None

    SHOpenFolderAndSelectItems = windll.shell32.SHOpenFolderAndSelectItems
    SHOpenFolderAndSelectItems.argtypes = [
        PCIDLIST_ABSOLUTE,
        wintypes.UINT,
        PCUITEMID_CHILD_ARRAY,
        DWORD,
    ]
    SHOpenFolderAndSelectItems.restype = HRESULT

    SFGAOF = wintypes.ULONG

    SHParseDisplayName = windll.shell32.SHParseDisplayName
    SHParseDisplayName.argtypes = [
        wintypes.LPCWSTR,
        ctypes.POINTER(IBindCtx),
        ctypes.POINTER(PIDLIST_ABSOLUTE),
        SFGAOF,
        ctypes.POINTER(SFGAOF),
    ]
    SHParseDisplayName.restype = HRESULT

    @enum
    class FOLDERID(str):
        LINKS = "{bfb9d5e0-c6a9-404c-b2b2-ae6db6af4968}"

    @enum
    class SHGFPType(int):
        CURRENT = 0
        DEFAULT = 1

    @enum
    class CSIDL(int):
        DESKTOP = 0x0000
        PERSONAL = 0x0005
        APPDATA = 0x001A
        MYMUSIC = 0x000D
        PROFILE = 0x0028

    @enum
    class CSIDLFlag(int):
        PER_USER_INIT = 0x0800
        NO_ALIAS = 0x1000
        DONT_UNEXPAND = 0x2000
        DONT_VERIFY = 0x4000
        CREATE = 0x8000
        MASK = 0xFF00

    @enum
    class KnownFolderFlag(int):
        SIMPLE_IDLIST = 0x00000100
        NOT_PARENT_RELATIVE = 0x00000200
        DEFAULT_PATH = 0x00000400
        INIT = 0x00000800
        NO_ALIAS = 0x00001000
        DONT_UNEXPAND = 0x00002000
        DONT_VERIFY = 0x00004000
        CREATE = 0x00008000
        NO_APPCONTAINER_REDIRECTION = 0x00010000
        ALIAS_ONLY = 0x80000000



# Copyright 2004-2006 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012 Christoph Reiter
#           2013,16,21 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _
from quodlibet import app
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Icons
from quodlibet.util import is_linux, is_osx, print_w, print_d
from quodlibet.util.environment import dbus_name_owned

from .prefs import Preferences
from .systemtray import SystemTray


if is_osx():
    # Works, but not without problems:
    # https://github.com/quodlibet/quodlibet/issues/1870
    # The dock menu is more useful so disable.
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError


def get_indicator_impl():
    """Returns a BaseIndicator implementation depending on the environ"""

    use_app_indicator = is_linux() and dbus_name_owned("org.kde.StatusNotifierWatcher")

    print_d(f"use app indicator: {use_app_indicator}")
    if not use_app_indicator:
        return SystemTray
    try:
        from .appindicator import AppIndicator
    except ImportError as e:
        print_w(f"Loading AppIndicator failed ({e}). Using {SystemTray}")
        # no indicator, fall back
        return SystemTray
    else:
        return AppIndicator


class TrayIconPlugin(EventPlugin):
    PLUGIN_ID = "Tray Icon"
    PLUGIN_NAME = _("Tray Icon")
    PLUGIN_DESC = _("Controls Quod Libet from the system tray.")
    PLUGIN_ICON = Icons.USER_DESKTOP

    def enabled(self):
        impl = get_indicator_impl()
        self._tray = impl()
        self._tray.set_song(app.player.song)
        self._tray.set_info_song(app.player.info)
        self._tray.set_paused(app.player.paused)

    def disabled(self):
        self._tray.remove()
        del self._tray

    def PluginPreferences(self, parent):
        return Preferences()

    def plugin_on_song_started(self, song):
        self._tray.set_song(app.player.song)
        self._tray.set_info_song(app.player.info)

    def plugin_on_paused(self):
        self._tray.set_paused(True)

    def plugin_on_unpaused(self):
        self._tray.set_paused(False)



# Copyright 2017 Didier Villevalois,
#           2010 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gdk, Gtk

from quodlibet import _
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons
from quodlibet.qltk.window import Dialog


class TapBpmPanel(Gtk.VBox):
    def __init__(self, parent, song):
        super().__init__()

        self.dialog = parent
        self.song = song
        self.original_bpm = song["bpm"] if "bpm" in song else _("n/a")

        self.set_margin_bottom(6)
        self.set_spacing(6)

        box = Gtk.HBox()
        box.set_spacing(6)
        # TRANSLATORS: BPM mean "beats per minute"
        box.pack_start(Gtk.Label(_("BPM:")), False, True, 0)
        self.bpm_label = Gtk.Label(_("n/a"))
        self.bpm_label.set_xalign(0.5)
        box.pack_start(self.bpm_label, True, True, 0)

        self.reset_btn = Gtk.Button(label=_("Reset"))
        self.reset_btn.connect("clicked", lambda *x: self.reset())
        box.pack_end(self.reset_btn, False, True, 0)

        self.pack_start(box, False, True, 0)

        self.tap_btn = Gtk.Button(label=_("Tap"))
        self.tap_btn.connect("button-press-event", self.tap)
        self.tap_btn.connect("key-press-event", self.key_tap)
        self.pack_start(self.tap_btn, True, True, 0)

        self.init_tap()
        self.update()

        self.show_all()

    def update(self):
        has_new_bpm = self.clicks > 1

        self.dialog.set_response_sensitive(Gtk.ResponseType.OK, has_new_bpm)
        self.reset_btn.set_sensitive(has_new_bpm)

        if self.clicks > 1:
            self.bpm_label.set_text(f"{self.floating_bpm:.1f}")
        elif self.clicks == 1:
            self.bpm_label.set_text(_("n/a"))
        else:
            self.bpm_label.set_text(str(self.original_bpm))

        # Give focus back to the tap button even if reset was pressed
        self.tap_btn.grab_focus()

    def tap(self, widget, event):
        self.count_tap(event.time)
        self.update()

    def key_tap(self, widget, event):
        if event.keyval != Gdk.KEY_space and event.keyval != Gdk.KEY_Return:
            return False

        self.count_tap(event.time)
        self.update()
        return True

    def reset(self):
        self.init_tap()
        self.update()

    def save(self):
        self.song["bpm"] = f"{self.floating_bpm:.0f}"
        self.song.write()

    def init_tap(self):
        self.bpm = 0.0
        self.clicks = 0
        self.last = 0
        self.last_times = []
        self.last_bpms = []
        self.last_floating_bpms = []
        self.last_floating_squares = []
        self.bpms_sum = 0.0
        self.squares_sum = 0.0
        self.average_count = 0
        self.min = 0
        self.max = 0
        try:
            self.floating_bpm = float(self.original_bpm)
        except ValueError:
            self.floating_bpm = 0.0
        self.floating_square = 0.0
        self.keep = 100

        self.pause = 3
        self.min_weight = 0.01

    def count_tap(self, now):
        now = now / 1000.0
        # reset?
        if now - self.last > self.pause:
            self.clicks = 0
            self.bpm = 0.0

            self.last_times = []
            self.last_bpms = []
            self.last_floating_bpms = []
            self.last_floating_squares = []
            self.bpms_sum = 0.0
            self.squares_sum = 0.0
            self.average_count = 0
            self.min = 0
            self.max = 0
            self.floating_bpm = 0.0
            self.floating_square = 0.0
        elif now > self.last:
            # Use previous 5 values to average BPM
            bpms = []
            bpms.append(60.0 / (now - self.last))
            # and four out of the list
            for i in iter(range(1, 5)):
                if len(self.last_times) <= i:
                    break
                bpms.append((i + 1) * 60.0 / (now - self.last_times[-i]))

            self.bpm = sum(bpms) / len(bpms)

            # Exponentially weighted floating average
            weight = (1.0 / self.clicks) ** 0.5
            if weight < self.min_weight:
                weight = self.min_weight
            self.floating_bpm = self.floating_bpm * (1.0 - weight) + self.bpm * weight
            self.floating_square = (
                self.floating_square * (1.0 - weight) + self.bpm * self.bpm * weight
            )

            if self.bpm < self.min or self.average_count == 0:
                self.min = self.bpm
            if self.bpm > self.max or self.average_count == 0:
                self.max = self.bpm
            self.bpms_sum += self.bpm
            self.squares_sum += self.bpm * self.bpm
            self.average_count += 1

            # Update history
            self.last_times = self.last_times[-(self.keep - 1) :] + [self.last]
            self.last_bpms = self.last_bpms[-(self.keep - 1) :] + [self.bpm]
            self.last_floating_bpms = self.last_floating_bpms[-(self.keep - 1) :] + [
                self.floating_bpm
            ]
            self.last_floating_squares = self.last_floating_squares[
                -(self.keep - 1) :
            ] + [self.floating_square]

        self.last = now
        self.clicks += 1


class TapBpm(SongsMenuPlugin):
    PLUGIN_ID = "Tap BPM"
    PLUGIN_NAME = _("Tap BPM")
    PLUGIN_DESC = _("🥁 Tap BPM for the selected song.")
    PLUGIN_ICON = Icons.EDIT
    PLUGIN_VERSION = "0.1"

    def plugin_song(self, song):
        self._window = window = Dialog(title=_("Tap BPM"), parent=self.plugin_window)

        window.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        window.add_icon_button(_("_Save"), Icons.DOCUMENT_SAVE, Gtk.ResponseType.OK)

        window.set_default_size(300, 100)
        window.set_border_width(6)
        self.__resp_sig = window.connect("response", self.response)

        self._panel = TapBpmPanel(window, song)
        window.vbox.pack_start(self._panel, False, True, 0)

        window.vbox.show_all()
        window.present()

    def response(self, win, response):
        if response == Gtk.ResponseType.OK:
            # Save metadata
            self._panel.save()

        win.hide()
        win.disconnect(self.__resp_sig)
        win.destroy()



# Copyright 2015 Christoph Reiter
#           2017 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

from quodlibet.qltk import Icons

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

import fcntl
import subprocess

from gi.repository import GLib, GObject

from quodlibet import _
from quodlibet.plugins.events import EventPlugin
from quodlibet.util import print_d
from quodlibet import app


def get_headphone_status():
    """Returns if headphones are connected."""

    # No idea if this is a good way... but works here :)
    try:
        data = subprocess.check_output(["pactl", "list", "sinks"])
    except OSError:
        return False
    for line in data.splitlines():
        if line.strip() == b"Active Port: analog-output-headphones":
            return True
    else:
        return False


class HeadphoneAction:
    DISCONNECTED = 0
    CONNECTED = 1


class HeadphoneMonitor(GObject.Object):
    """Monitors the headphone connection state advertised by pulseaudio.

    After start() is called will emit a signal in case headphones
    get connected or disconnected.

    If pulseaudio isn't active this will work but always return
    a disconnected status.

    The changed signal will never be emitted with the same status multiple
    times.
    """

    __gsignals__ = {
        "action": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    def __init__(self):
        super().__init__()
        self._subscribe_id = None
        self._process = None
        self._status = None

    def is_connected(self):
        """Returns whether headphones are currently connected"""

        if self._status is None:
            raise Exception("call start() first")

        return self._status

    def _emit(self):
        self.emit(
            "action",
            HeadphoneAction.CONNECTED if self._status else HeadphoneAction.DISCONNECTED,
        )

    def _update_status(self):
        assert self._status is not None
        new_status = get_headphone_status()
        if new_status != self._status:
            self._status = new_status
            self._emit()
            return

    def start(self):
        """Start the monitoring process.

        Once this gets called the "changed" signal will be emitted.
        """

        null = open(os.devnull, "wb")
        try:
            self._process = subprocess.Popen(
                ["pactl", "subscribe"], stdout=subprocess.PIPE, stderr=null
            )
        except OSError:
            self._status = False
            return

        f = self._process.stdout
        fcntl.fcntl(f, fcntl.F_SETFL, os.O_NONBLOCK)

        def can_read_cb(fd, flags):
            if flags & (GLib.IOCondition.HUP | GLib.IOCondition.ERR):
                f.close()
                self._subscribe_id = None
                return False

            data = f.read()
            if not data:
                f.close()
                self._subscribe_id = None
                return False

            # querying the status also results in client events which would
            # lead us into an endless loop. Instead just something if there
            # is a sink event
            if b" on sink " in data:
                self._update_status()
            return True

        self._status = get_headphone_status()
        self._subscribe_id = GLib.io_add_watch(
            f,
            GLib.PRIORITY_HIGH,
            GLib.IOCondition.IN | GLib.IOCondition.ERR | GLib.IOCondition.HUP,
            can_read_cb,
        )

    def stop(self):
        """Stop the monitoring process.

        After this returns no signal will be emitted.
        Can be called multiple times.
        start() can be called to start monitoring again after this returns.
        """

        if self._subscribe_id is not None:
            GLib.source_remove(self._subscribe_id)
            self._subscribe_id = None

        if self._process is not None:
            self._process.terminate()
            self._process.wait()
            self._process = None

        self._status = None


class HeadphoneMonitorPlugin(EventPlugin):
    PLUGIN_ID = "HeadphoneMonitor"
    PLUGIN_NAME = _("Pause on Headphone Unplug")
    PLUGIN_DESC = _(
        "Pauses in case headphones get unplugged and unpauses in "
        "case they get plugged in again."
    )
    PLUGIN_ICON = Icons.MEDIA_PLAYBACK_PAUSE

    def enabled(self):
        self._was_paused = False
        self._do_act = False
        self._mon = HeadphoneMonitor()
        self._mon.connect("action", self._changed)
        self._mon.start()

    def _changed(self, mon, action):
        if action == HeadphoneAction.DISCONNECTED:
            print_d("Headphones disconnected")
            if self._do_act:
                do_act = self._do_act
                self._was_paused = app.player.paused
                app.player.paused = True
                self._do_act = do_act
        elif action == HeadphoneAction.CONNECTED:
            print_d("Headphones connected")
            if self._do_act:
                do_act = self._do_act
                app.player.paused = self._was_paused
                self._do_act = do_act

    def disabled(self):
        self._mon.stop()
        del self._mon

    def plugin_on_paused(self):
        self._do_act = False

    def plugin_on_unpaused(self):
        self._do_act = self._mon.is_connected()



# Copyright 2012 Christoph Reiter
#           2018 Ludovic Druette
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, GObject, Gst

from quodlibet import _
from quodlibet.plugins import PluginImportError
from quodlibet.plugins.gstelement import GStreamerPlugin
from quodlibet.qltk.util import GSignals
from quodlibet import qltk
from quodlibet import config


_PLUGIN_ID = "pitch"

_SETTINGS = {
    "rate": [_("R_ate:"), 1.0],
    "tempo": [_("_Tempo:"), 1.0],
    "pitch": [_("_Pitch:"), 1.0],
}


def get_cfg(option):
    cfg_option = f"{_PLUGIN_ID}_{option}"
    default = _SETTINGS[option][1]

    if option == "rate":
        return config.getfloat("plugins", cfg_option, default)
    if option == "tempo":
        return config.getfloat("plugins", cfg_option, default)
    if option == "pitch":
        return config.getfloat("plugins", cfg_option, default)
    return None


def set_cfg(option, value):
    cfg_option = f"{_PLUGIN_ID}_{option}"
    if get_cfg(option) != value:
        config.set("plugins", cfg_option, value)


class Preferences(Gtk.VBox):
    __gsignals__: GSignals = {
        "changed": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        super().__init__(spacing=12)

        table = Gtk.Table(n_rows=3, n_columns=3)
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        labels = {}
        for idx, key in enumerate(["tempo", "rate", "pitch"]):
            label = Gtk.Label(label=_SETTINGS[key][0])
            labels[key] = label
            label.set_alignment(0.0, 0.5)
            label.set_padding(0, 6)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )

        def scale_changed(scale, option):
            value = scale.get_value()
            set_cfg(option, value)
            self.emit("changed")

        for idx, key in enumerate(["tempo", "rate", "pitch"]):
            adjustment = Gtk.Adjustment(0, 0.1, 3, 0.1, 1, 0)
            scale = Gtk.HScale(adjustment=adjustment)
            scale.set_digits(2)
            scale.add_mark(1.0, Gtk.PositionType.BOTTOM, None)
            labels[key].set_mnemonic_widget(scale)
            scale.set_draw_value(False)
            table.attach(scale, 1, 2, idx, idx + 1)
            spin = Gtk.SpinButton(adjustment=adjustment)
            spin.set_digits(2)
            table.attach(
                spin,
                2,
                3,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            scale.connect("value-changed", scale_changed, key)
            scale.set_value(get_cfg(key))

        self.pack_start(qltk.Frame(_("Preferences"), child=table), True, True, 0)


class Pitch(GStreamerPlugin):
    PLUGIN_ID = _PLUGIN_ID
    PLUGIN_NAME = _("Audio Pitch / Speed")
    PLUGIN_DESC = _("Controls the pitch of an audio stream.")

    @classmethod
    def setup_element(cls):
        return Gst.ElementFactory.make("pitch", cls.PLUGIN_ID)

    @classmethod
    def update_element(cls, element):
        for key in ["tempo", "rate", "pitch"]:
            element.set_property(key, get_cfg(key))

    @classmethod
    def PluginPreferences(cls, window):
        prefs = Preferences()
        prefs.connect("changed", lambda *x: cls.queue_update())
        return prefs


if not Pitch.setup_element():
    raise PluginImportError("GStreamer element 'pitch' missing (gst-plugins-bad)")



# Copyright 2005 Joe Wreschnig
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from gi.repository import Gtk

from quodlibet.util import connect_obj
from quodlibet.util.misc import total_ordering


@total_ordering
class RenameFilesPlugin:
    """Plugins of this type must subclass a GTK widget. They will be
    packed into the RenameFiles pane (currently a ScrolledWindow hidden
    with an expander, but that might change).

    The 'filter' function will be called with the song's original filename
    as a string (probably in the local filesystem encoding) and the proposed
    new filename as a unicode object. It should return an
    appropriate-transformed filename, still as a unicode object.

    The plugin must provide either a 'changed' or 'preview'. 'preview'
    causes the entire display to be re-previewed. 'changed' causes the
    Preview button to made sensitive, and Save to be disabled.

    If the 'active' attribute is false, the filter will not be called.
    This is particularly useful for gtk.CheckButtons.

    The '_order' attribute decides the sort order of the plugin. The
    default filters have orders between 1 and 2. Plugins have order 0 by
    default. Plugins with equal orders are sorted by class name."""

    _order = 0
    active = False

    def filter(self, original_filename, value):
        return value

    def filter_list(self, origs, names):
        return list(map(self.filter, origs, names))

    def __eq__(self, other):
        return (self._order, type(self).__name__) == (
            other._order,
            type(other).__name__,
        )

    def __lt__(self, other):
        return (self._order, type(self).__name__) < (other._order, type(other).__name__)


@total_ordering
class TagsFromPathPlugin:
    """Plugins of this type must subclass a GTK widget. They will be
    packed into the TagsFromPath pane (currently a ScrolledWindow hidden
    with an expander, but that might change).

    The 'filter' function will be called with the tag and proposed value
    as a unicode object. It should return an appropriate-transformed
    filename, still as a unicode object.

    If you need to work on a set of filenames at once, you should
    instead overload the 'filter_list' function, which takes two lists;
    one of original filenames, the other of proposed new filenames.
    The default filter_list function calls filter on each item.

    The plugin must provide either a 'changed' or 'preview'. 'preview'
    causes the entire display to be re-previewed. 'changed' causes the
    Preview button to made sensitive, and Save to be disabled.

    If the 'active' attribute is false, the filter will not be called.
    This is particularly useful for gtk.CheckButtons.

    The '_order' attribute decides the sort order of the plugin. The
    default filters have orders between 1 and 2. Plugins have order 0 by
    default. Plugins with equal orders are sorted by class name."""

    _order = 0
    active = False

    def filter(self, tag, value):
        return value

    def __eq__(self, other):
        return (self._order, type(self).__name__) == (
            other._order,
            type(other).__name__,
        )

    def __lt__(self, other):
        return (self._order, type(self).__name__) < (other._order, type(other).__name__)


class EditTagsPlugin(Gtk.ImageMenuItem):
    """Plugins of this type are subclasses of gtk.ImageMenuItem.
    They will be added to the context menu of the EditTags tree view.

    The 'tags' attribute is a list of tags this plugin should appear on,
    or false if it should appear for all tags. This must be a class
    attribute, as it is checked before instantiation.

    The 'needs' attribute is a list of tags that must be editable in
    the currently selected songs for the plugin to be sensitive.

    The constructor is called with the tag and value for that row. This
    can be used to set the sensitivity of the menu item, or change its
    text.

    When clicked, the 'activated' function is called on the object,
    again with the tag name and value. It should return a list of
    (tag, value) tuples to replace the previous tag/value with.

    The '_order' attribute decides the sort order of the plugin. The
    default items have orders between 0 and 1. Plugins have order 2.0 by
    default. Plugins with equal orders are sorted by class name.

    How to Handle Submenus
    ----------------------
    If the menu item is given a submenu, magic happens. In particular,
    the callbacks are proxied to the submenu's items, and are attached,
    via connect_object, to make sure activated is called on the original
    item.

    So, to handle submenus,
      1. Make a submenu and attach it to your menu item.2
      2. To each item in the submenu, attach its 'activate' signal to
         something appropriate to prepare the original item's
         activated method,
      3. Because that method will be called after the subitem is
         clicked on, and your activate handler runs.
    """

    tags: list[str] = []
    needs: list[str] = []
    _order = 2.0

    def activated(self, tag, value):
        return [(tag, value)]

    def connect(self, signal, callback, *args, **kwargs):
        if self.get_submenu():
            for item in self.get_submenu().get_children():
                connect_obj(item, signal, callback, self, *args, **kwargs)
        else:
            super().connect(signal, callback, *args, **kwargs)



# Copyright 2012,2013 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys

if os.name == "nt" or sys.platform == "darwin":
    from quodlibet.plugins import PluginNotSupportedError

    raise PluginNotSupportedError

from gi.repository import Gtk, GdkPixbuf
from senf import fsn2uri

import dbus
import dbus.service

from quodlibet import _
from quodlibet import app
from quodlibet.plugins.events import EventPlugin
from quodlibet.pattern import Pattern
from quodlibet.qltk import Icons
from quodlibet.util.dbusutils import DBusIntrospectable, DBusProperty
from quodlibet.util.dbusutils import dbus_unicode_validate as unival
from quodlibet.util import NamedTemporaryFile

BASE_PATH = "/org/gnome/UPnP/MediaServer2"
BUS_NAME = "org.gnome.UPnP.MediaServer2.QuodLibet"


class MediaServer(EventPlugin):
    PLUGIN_ID = "mediaserver"
    PLUGIN_NAME = _("UPnP AV Media Server")
    PLUGIN_DESC = _(
        "Exposes all albums to the Rygel UPnP Media Server "
        "through the MediaServer2 D-Bus interface."
    )
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP

    def PluginPreferences(self, parent):
        vbox = Gtk.VBox(spacing=12)

        conf_exp = _(
            "Ensure the following is in your rygel config file "
            "(~/.config/rygel.conf):"
        )
        conf_cont = (
            "[External]\n"
            "enabled=true\n\n"
            "[org.gnome.UPnP.MediaServer2.QuodLibet]\n"
            "enabled=true"
        )

        exp_lbl = Gtk.Label(label=conf_exp)
        exp_lbl.set_selectable(True)
        exp_lbl.set_line_wrap(True)
        exp_lbl.set_alignment(0, 0)

        conf_lbl = Gtk.Label()
        conf_lbl.set_selectable(True)
        conf_lbl.set_alignment(0, 0)
        conf_lbl.set_markup(f"<span font='mono'>{conf_cont}</span>")

        vbox.pack_start(exp_lbl, True, False, 0)
        vbox.pack_start(conf_lbl, True, False, 0)
        return vbox

    def enabled(self):
        try:
            dbus.SessionBus()
        except dbus.DBusException:
            self.objects = []
            return

        entry = EntryObject()
        albums = AlbumsObject(entry, app.library)
        song = SongObject(app.library, [albums])
        icon = Icon(entry)

        self.objects = [entry, albums, song, icon]

    def disabled(self):
        for obj in self.objects:
            obj.remove_from_connection()

        for obj in self.objects:
            obj.destroy()

        del self.objects

        import gc

        gc.collect()


class DBusPropertyFilter(DBusProperty):
    """Adds some methods to support the MediaContainer property filtering."""

    def get_properties_for_filter(self, interface, filter_):
        props = self.get_properties(interface)
        if "*" not in filter_:
            props = [p for p in props if p[1] in filter_]
        return props

    def get_values(self, properties, path="/"):
        result = {}
        for iface, prop in properties:
            result[prop] = self.get_value(iface, prop, path)
        return result


class MediaContainer:
    IFACE = "org.gnome.UPnP.MediaContainer2"
    ISPEC_PROP = """
<property type="u" name="ChildCount" access="read"/>
<property type="u" name="ItemCount" access="read"/>
<property type="u" name="ContainerCount" access="read"/>
<property type="b" name="Searchable" access="read"/>
<property type="o" name="Icon" access="read"/>
"""
    ISPEC = """
<method name="ListChildren">
    <arg type="u" name="offset" direction="in"/>
    <arg type="u" name="max" direction="in"/>
    <arg type="as" name="filter" direction="in"/>
    <arg type="aa{sv}" name="arg_3" direction="out"/>
</method>
<method name="ListContainers">
    <arg type="u" name="offset" direction="in"/>
    <arg type="u" name="max" direction="in"/>
    <arg type="as" name="filter" direction="in"/>
    <arg type="aa{sv}" name="arg_3" direction="out"/>
</method>
<method name="ListItems">
    <arg type="u" name="offset" direction="in"/>
    <arg type="u" name="max" direction="in"/>
    <arg type="as" name="filter" direction="in"/>
    <arg type="aa{sv}" name="arg_3" direction="out"/>
</method>
<method name="SearchObjects">
    <arg type="s" name="query" direction="in"/>
    <arg type="u" name="offset" direction="in"/>
    <arg type="u" name="max" direction="in"/>
    <arg type="as" name="filter" direction="in"/>
    <arg type="aa{sv}" name="arg_4" direction="out"/>
</method>

<signal name="Updated"/>
"""

    def __init__(self, optional=()):
        self.set_introspection(MediaContainer.IFACE, MediaContainer.ISPEC)

        props = ["ChildCount", "ItemCount", "ContainerCount", "Searchable"]
        props += list(optional)
        self.set_properties(MediaContainer.IFACE, MediaContainer.ISPEC_PROP, wl=props)

        self.implement_interface(MediaContainer.IFACE, MediaObject.IFACE)

    def emit_updated(self, path="/"):
        self.Updated(rel=path)

    @dbus.service.method(
        IFACE, in_signature="uuas", out_signature="aa{sv}", rel_path_keyword="path"
    )
    def ListChildren(self, offset, max_, filter_, path):
        if self.SUPPORTS_MULTIPLE_OBJECT_PATHS:
            return self.list_children(offset, max_, filter_, path)
        return self.list_children(offset, max_, filter_)

    @dbus.service.method(
        IFACE, in_signature="uuas", out_signature="aa{sv}", rel_path_keyword="path"
    )
    def ListContainers(self, offset, max_, filter_, path):
        if self.SUPPORTS_MULTIPLE_OBJECT_PATHS:
            return self.list_containers(offset, max_, filter_, path)
        return self.list_containers(offset, max_, filter_)

    @dbus.service.method(
        IFACE, in_signature="uuas", out_signature="aa{sv}", rel_path_keyword="path"
    )
    def ListItems(self, offset, max_, filter_, path):
        if self.SUPPORTS_MULTIPLE_OBJECT_PATHS:
            return self.list_items(offset, max_, filter_, path)
        return self.list_items(offset, max_, filter_)

    @dbus.service.method(
        IFACE, in_signature="suuas", out_signature="aa{sv}", rel_path_keyword="path"
    )
    def SearchObjects(self, query, offset, max_, filter_, path):
        return []

    @dbus.service.signal(IFACE, rel_path_keyword="rel")
    def Updated(self, rel=""):
        pass


class MediaObject:
    IFACE = "org.gnome.UPnP.MediaObject2"
    ISPEC = """
<property type="o" name="Parent" access="read"/>
<property type="s" name="Type" access="read"/>
<property type="o" name="Path" access="read"/>
<property type="s" name="DisplayName" access="read"/>
"""
    parent = None

    def __init__(self, parent=None):
        self.set_properties(MediaObject.IFACE, MediaObject.ISPEC)
        self.parent = parent or self


class MediaItem:
    IFACE = "org.gnome.UPnP.MediaItem2"
    ISPEC = """
<property type="as" name="URLs" access="read"/>
<property type="s" name="MIMEType" access="read"/>

<property type="x" name="Size" access="read"/>
<property type="s" name="Artist" access="read"/>
<property type="s" name="Album" access="read"/>
<property type="s" name="Date" access="read"/>
<property type="s" name="Genre" access="read"/>
<property type="s" name="DLNAProfile" access="read"/>

<property type="i" name="Duration" access="read"/>
<property type="i" name="Bitrate" access="read"/>
<property type="i" name="SampleRate" access="read"/>
<property type="i" name="BitsPerSample" access="read"/>

<property type="i" name="Width" access="read"/>
<property type="i" name="Height" access="read"/>
<property type="i" name="ColorDepth" access="read"/>
<property type="i" name="PixelWidth" access="read"/>
<property type="i" name="PixelHeight" access="read"/>
<property type="o" name="Thumbnail" access="read"/>

<property type="o" name="AlbumArt" access="read"/>

<property type="i" name="TrackNumber" access="read"/>
"""

    def __init__(self, optional=()):
        props = ["URLs", "MIMEType"] + list(optional)
        self.set_properties(MediaItem.IFACE, MediaItem.ISPEC, wl=props)
        self.implement_interface(MediaItem.IFACE, MediaObject.IFACE)


class EntryObject(
    MediaContainer,
    MediaObject,
    DBusPropertyFilter,
    DBusIntrospectable,
    dbus.service.Object,
):
    PATH = BASE_PATH + "/QuodLibet"
    DISPLAY_NAME = "@REALNAME@'s Quod Libet on @HOSTNAME@"

    def __init__(self):
        self.__sub = []

        DBusIntrospectable.__init__(self)
        DBusPropertyFilter.__init__(self)
        MediaObject.__init__(self)
        MediaContainer.__init__(self, optional=["Icon"])

        bus = dbus.SessionBus()
        name = dbus.service.BusName(BUS_NAME, bus)
        dbus.service.Object.__init__(self, bus, self.PATH, name)

    def get_property(self, interface, name):
        if interface == MediaContainer.IFACE:
            if name == "ChildCount":
                return len(self.__sub)
            if name == "ItemCount":
                return 0
            if name == "ContainerCount":
                return len(self.__sub)
            if name == "Searchable":
                return False
            if name == "Icon":
                return Icon.PATH
        elif interface == MediaObject.IFACE:
            if name == "Parent":
                return self.parent.PATH
            if name == "Type":
                return "container"
            if name == "Path":
                return self.PATH
            if name == "DisplayName":
                return self.DISPLAY_NAME
        return None

    def destroy(self):
        # break cycle
        del self.__sub
        del self.parent

    def register_child(self, child):
        self.__sub.append(child)
        self.emit_properties_changed(
            MediaContainer.IFACE, ["ChildCount", "ContainerCount"]
        )

    def list_containers(self, offset, max_, filter_):
        props = self.get_properties_for_filter(MediaContainer.IFACE, filter_)
        end = (max_ and offset + max_) or None

        result = []
        for sub in self.__sub[offset:end]:
            result.append(sub.get_values(props))
        return result

    list_children = list_containers

    def list_items(self, offset, max_, filter_):
        return []


SUPPORTED_SONG_PROPERTIES = (
    "Size",
    "Artist",
    "Album",
    "Date",
    "Genre",
    "Duration",
    "TrackNumber",
)


class DummySongObject(MediaItem, MediaObject, DBusPropertyFilter, DBusIntrospectable):
    """A dummy song object that is not exported on the bus, but supports
    the usual interfaces.

    You need to assign a real song before using it, and have to pass
    a path prefix.

    The path of the song is /org/gnome/UPnP/MediaServer2/Song/<PREFIX>/SongID
    This lets us reconstruct the original parent path:
    /org/gnome/UPnP/MediaServer2/<PREFIX>

    atm. a prefix can look like "Albums/123456"
    """

    SUPPORTS_MULTIPLE_OBJECT_PATHS = False
    __pattern = Pattern("<discnumber|<discnumber>.><tracknumber>. <title>")

    def __init__(self, parent):
        DBusIntrospectable.__init__(self)
        DBusPropertyFilter.__init__(self)
        MediaObject.__init__(self, parent)
        MediaItem.__init__(self, optional=SUPPORTED_SONG_PROPERTIES)

    def set_song(self, song, prefix):
        self.__song = song
        self.__prefix = prefix

    def get_property(self, interface, name):
        if interface == MediaObject.IFACE:
            if name == "Parent":
                return BASE_PATH + "/" + self.__prefix
            if name == "Type":
                return "music"
            if name == "Path":
                path = SongObject.PATH
                path += "/" + self.__prefix + "/" + str(id(self.__song))
                return path
            if name == "DisplayName":
                return unival(self.__song.comma("title"))
        elif interface == MediaItem.IFACE:
            if name == "URLs":
                return [self.__song("~uri")]
            if name == "MIMEType":
                mimes = self.__song.mimes
                return mimes and mimes[0]
            if name == "Size":
                return self.__song("~#filesize")
            if name == "Artist":
                return unival(self.__song.comma("artist"))
            if name == "Album":
                return unival(self.__song.comma("album"))
            if name == "Date":
                return unival(self.__song.comma("date"))
            if name == "Genre":
                return unival(self.__song.comma("genre"))
            if name == "Duration":
                return self.__song("~#length")
            if name == "TrackNumber":
                return self.__song("~#track", 0)
        return None


class DummyAlbumObject(
    MediaContainer, MediaObject, DBusPropertyFilter, DBusIntrospectable
):
    SUPPORTS_MULTIPLE_OBJECT_PATHS = False
    __pattern = Pattern("<albumartist|<~albumartist~album>|<~artist~album>>")

    def __init__(self, parent):
        DBusIntrospectable.__init__(self)
        DBusPropertyFilter.__init__(self)
        MediaObject.__init__(self, parent)
        MediaContainer.__init__(self)
        self.__song = DummySongObject(self)

    def get_dummy(self, song):
        self.__song.set_song(song, "Albums/" + str(id(self.__album)))
        return self.__song

    def set_album(self, album):
        self.__album = album
        self.PATH = self.parent.PATH + "/" + str(id(album))

    def get_property(self, interface, name):
        if interface == MediaContainer.IFACE:
            if name == "ChildCount" or name == "ItemCount":
                return len(self.__album.songs)
            if name == "ContainerCount":
                return 0
            if name == "Searchable":
                return False
        elif interface == MediaObject.IFACE:
            if name == "Parent":
                return self.parent.PATH
            if name == "Type":
                return "container"
            if name == "Path":
                return self.PATH
            if name == "DisplayName":
                return unival(self.__pattern % self.__album)
        return None

    def list_containers(self, offset, max_, filter_):
        return []

    def list_items(self, offset, max_, filter_):
        songs = sorted(self.__album.songs, key=lambda s: s.sort_key)
        dummy = self.get_dummy(None)
        props = dummy.get_properties_for_filter(MediaItem.IFACE, filter_)
        end = (max_ and offset + max_) or None

        result = []
        for song in songs[offset:end]:
            result.append(self.get_dummy(song).get_values(props))
        return result

    list_children = list_items


class SongObject(
    MediaItem,
    MediaObject,
    DBusProperty,
    DBusIntrospectable,
    dbus.service.FallbackObject,
):
    PATH = BASE_PATH + "/Song"

    def __init__(self, library, users):
        DBusIntrospectable.__init__(self)
        DBusProperty.__init__(self)
        MediaObject.__init__(self, None)
        MediaItem.__init__(self, optional=SUPPORTED_SONG_PROPERTIES)

        bus = dbus.SessionBus()
        self.ref = dbus.service.BusName(BUS_NAME, bus)
        dbus.service.FallbackObject.__init__(self, bus, self.PATH)

        self.__library = library
        self.__map = {id(v): v for v in self.__library.values()}
        self.__reverse = {v: k for k, v in self.__map.items()}

        self.__song = DummySongObject(self)

        self.__users = users

        signals = [
            ("changed", self.__songs_changed),
            ("removed", self.__songs_removed),
            ("added", self.__songs_added),
        ]
        self.__sigs = (self.__library.connect(x[0], x[1]) for x in signals)

    def __songs_changed(self, lib, songs):
        # We don't know what changed, so get all properties
        props = [p[1] for p in self.get_properties(MediaItem.IFACE)]

        for song in songs:
            song_id = str(id(song))
            # https://github.com/quodlibet/quodlibet/issues/id=1127
            # XXX: Something is emitting wrong changed events..
            # ignore song_ids we don't know for now
            if song_id not in self.__map:
                continue
            for user in self.__users:
                # ask the user for the prefix with which the song is used
                prefix = user.get_prefix(song)
                path = "/" + prefix + "/" + song_id
                self.emit_properties_changed(MediaItem.IFACE, props, path)

    def __songs_added(self, lib, songs):
        for song in songs:
            new_id = id(song)
            self.__map[new_id] = song
            self.__reverse[song] = new_id

    def __songs_removed(self, lib, songs):
        for song in songs:
            del self.__map[self.__reverse[song]]
            del self.__reverse[song]

    def destroy(self):
        for signal_id in self.__sigs:
            self.__library.disconnect(signal_id)

    def get_dummy(self, song, prefix):
        self.__song.set_song(song, prefix)
        return self.__song

    def get_property(self, interface, name, path):
        # extract the prefix
        prefix, song_id = path[1:].rsplit("/", 1)
        song = self.__map[int(song_id)]
        return self.get_dummy(song, prefix).get_property(interface, name)


class AlbumsObject(
    MediaContainer,
    MediaObject,
    DBusPropertyFilter,
    DBusIntrospectable,
    dbus.service.FallbackObject,
):
    PATH = BASE_PATH + "/Albums"
    DISPLAY_NAME = "Albums"

    def __init__(self, parent, library):
        DBusIntrospectable.__init__(self)
        DBusPropertyFilter.__init__(self)
        MediaObject.__init__(self, parent)
        MediaContainer.__init__(self)

        bus = dbus.SessionBus()
        self.ref = dbus.service.BusName(BUS_NAME, bus)
        dbus.service.FallbackObject.__init__(self, bus, self.PATH)

        parent.register_child(self)

        self.__library = library.albums
        self.__library.load()

        self.__map = {id(v): v for v in self.__library.values()}
        self.__reverse = {v: k for k, v in self.__map.items()}

        signals = [
            ("changed", self.__albums_changed),
            ("removed", self.__albums_removed),
            ("added", self.__albums_added),
        ]
        self.__sigs = (self.__library.connect(x[0], x[1]) for x in signals)

        self.__dummy = DummyAlbumObject(self)

    def get_dummy(self, album):
        self.__dummy.set_album(album)
        return self.__dummy

    def get_path_dummy(self, path):
        return self.get_dummy(self.__map[int(path[1:])])

    def __albums_changed(self, lib, albums):
        for album in albums:
            rel_path = "/" + str(id(album))
            self.emit_updated(rel_path)
            self.emit_properties_changed(
                MediaContainer.IFACE,
                ["ChildCount", "ItemCount", "DisplayName"],
                rel_path,
            )

    def __albums_added(self, lib, albums):
        for album in albums:
            new_id = id(album)
            self.__map[new_id] = album
            self.__reverse[album] = new_id
        self.emit_updated()
        self.emit_properties_changed(
            MediaContainer.IFACE, ["ChildCount", "ContainerCount"]
        )

    def __albums_removed(self, lib, albums):
        for album in albums:
            del self.__map[self.__reverse[album]]
            del self.__reverse[album]
        self.emit_updated()
        self.emit_properties_changed(
            MediaContainer.IFACE, ["ChildCount", "ContainerCount"]
        )

    def get_prefix(self, song):
        album = self.__library[song.album_key]
        return "Albums/" + str(id(album))

    def destroy(self):
        for signal_id in self.__sigs:
            self.__library.disconnect(signal_id)

    def __get_albums_property(self, interface, name):
        if interface == MediaContainer.IFACE:
            if name == "ChildCount":
                return len(self.__library)
            if name == "ItemCount":
                return 0
            if name == "ContainerCount":
                return len(self.__library)
            if name == "Searchable":
                return False
        elif interface == MediaObject.IFACE:
            if name == "Parent":
                return self.parent.PATH
            if name == "Type":
                return "container"
            if name == "Path":
                return self.PATH
            if name == "DisplayName":
                return self.DISPLAY_NAME
        return None

    def get_property(self, interface, name, path):
        if path == "/":
            return self.__get_albums_property(interface, name)

        return self.get_path_dummy(path).get_property(interface, name)

    def __list_albums(self, offset, max_, filter_):
        props = self.get_properties_for_filter(MediaContainer.IFACE, filter_)
        albums = sorted(self.__library, key=lambda a: a.sort)
        end = (max_ and offset + max_) or None

        result = []
        for album in albums[offset:end]:
            result.append(self.get_dummy(album).get_values(props))
        return result

    def list_containers(self, offset, max_, filter_, path):
        if path == "/":
            return self.__list_albums(offset, max_, filter_)
        return []

    def list_items(self, offset, max_, filter_, path):
        if path != "/":
            return self.get_path_dummy(path).list_items(offset, max_, filter_)
        return []

    def list_children(self, offset, max_, filter_, path):
        if path == "/":
            return self.__list_albums(offset, max_, filter_)
        return self.get_path_dummy(path).list_children(offset, max_, filter_)


class Icon(
    MediaItem, MediaObject, DBusProperty, DBusIntrospectable, dbus.service.Object
):
    PATH = BASE_PATH + "/Icon"

    SIZE = 160

    def __init__(self, parent):
        DBusIntrospectable.__init__(self)
        DBusProperty.__init__(self)
        MediaObject.__init__(self, parent=parent)
        MediaItem.__init__(self, optional=["Height", "Width", "ColorDepth"])

        bus = dbus.SessionBus()
        name = dbus.service.BusName(BUS_NAME, bus)
        dbus.service.Object.__init__(self, bus, self.PATH, name)

        # https://bugzilla.gnome.org/show_bug.cgi?id=669677
        self.implement_interface("org.gnome.UPnP.MediaItem1", MediaItem.IFACE)

        # load into a pixbuf
        theme = Gtk.IconTheme.get_default()
        pixbuf = theme.load_icon(Icons.QUODLIBET, Icon.SIZE, 0)

        # make sure the size is right
        pixbuf = pixbuf.scale_simple(
            Icon.SIZE, Icon.SIZE, GdkPixbuf.InterpType.BILINEAR
        )
        self.__depth = pixbuf.get_bits_per_sample()

        # save and keep reference
        self.__f = f = NamedTemporaryFile()
        pixbuf.savev(f.name, "png", [], [])

    def get_property(self, interface, name):
        if interface == MediaObject.IFACE:
            if name == "Parent":
                return EntryObject.PATH
            if name == "Type":
                return "image"
            if name == "Path":
                return Icon.PATH
            if name == "DisplayName":
                return r"I'm an icon \o/"
        elif interface == MediaItem.IFACE:
            if name == "URLs":
                return [fsn2uri(self.__f.name)]
            if name == "MIMEType":
                return "image/png"
            if name == "Width" or name == "Height":
                return Icon.SIZE
            if name == "ColorDepth":
                return self.__depth
        return None

    def destroy(self):
        pass



# Copyright 2016 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import contextlib

import mutagen

from quodlibet import util
from quodlibet.util import print_w, print_d, reraise
from quodlibet.util.importhelper import load_dir_modules
from quodlibet.const import MinVersions


mimes = set()
"""A set of supported mime types"""

loaders = {}
"""A dict mapping file extensions to loaders (func returning an AudioFile)"""

types = set()
"""A set of AudioFile subclasses/implementations"""


class AudioFileError(Exception):
    """Base error for AudioFile, mostly IO/parsing related operations"""


class MutagenBug(AudioFileError):  # noqa
    """Raised in is caused by a mutagen bug, so we can highlight it"""


@contextlib.contextmanager
def translate_errors():
    """Context manager for mutagen calls to load/save. Translates exceptions
    to local ones.
    """

    try:
        yield
    except AudioFileError:
        raise
    except mutagen.MutagenError as e:
        reraise(AudioFileError, e)
    except Exception as e:
        reraise(MutagenBug, e)


def init():
    """Load/Import all formats.

    Before this is called loading a file and unpickling will not work.
    """

    global mimes, loaders, types

    MinVersions.MUTAGEN.check(mutagen.version)

    base = util.get_module_dir()
    formats = load_dir_modules(base, package=__package__)

    module_names = []
    for format in formats:
        name = format.__name__

        for ext in format.extensions:
            loaders[ext] = format.loader

        types.update(format.types)

        if format.extensions:
            for type_ in format.types:
                mimes.update(type_.mimes)
            module_names.append(name.split(".")[-1])

        # Migrate pre-0.16 library, which was using an undocumented "feature".
        sys.modules[name.replace(".", "/")] = format
        # Migrate old layout
        if name.startswith("quodlibet."):
            sys.modules[name.split(".", 1)[1]] = format

    # This can be used for the quodlibet.desktop file
    desktop_mime_types = (
        "MimeType=" + ";".join(sorted({m.split(";")[0] for m in mimes})) + ";"
    )
    print_d(desktop_mime_types)

    s = ", ".join(sorted(module_names))
    print_d(f"Supported formats: {s}")

    if not loaders:
        raise SystemExit("No formats found!")


def get_loader(filename):
    """Returns a callable which takes a filename and returns
    AudioFile or raises AudioFileError, or returns None.
    """

    ext = os.path.splitext(filename)[-1]
    return loaders.get(ext.lower())


def MusicFile(filename):  # noqa
    """Returns a AudioFile instance or None"""

    loader = get_loader(filename)
    if loader is not None:
        try:
            return loader(filename)
        except AudioFileError:
            print_w(f"Error loading {filename!r}")
            util.print_exc()
        except Exception:
            print_w(f"Error loading {filename!r}")
            raise
    return None


def filter(filename):
    """Returns True if the file extension is supported"""

    return get_loader(filename) is not None



# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Gdk, GObject

from quodlibet import config
from quodlibet import qltk
from quodlibet import _
from quodlibet.util import connect_obj, connect_destroy
from quodlibet.qltk.x import SymbolicIconImage, RadioMenuItem
from quodlibet.qltk.util import GSignals
from quodlibet.qltk.seekbutton import SeekButton
from quodlibet.util.dprint import print_e


class Volume(Gtk.VolumeButton):
    def __init__(self, player):
        super().__init__(size=Gtk.IconSize.MENU, use_symbolic=True)

        # https://bugzilla.gnome.org/show_bug.cgi?id=781605
        scales = qltk.find_widgets(self.get_popup(), Gtk.Scale)
        if scales:
            scales[0].props.round_digits = -1

        self.set_relief(Gtk.ReliefStyle.NORMAL)
        self.set_adjustment(Gtk.Adjustment.new(0, 0, 1, 0.05, 0.1, 0))

        popup = self.get_popup()
        if hasattr(Gtk, "Popover") and isinstance(popup, Gtk.Popover):
            popup.set_position(Gtk.PositionType.BOTTOM)

        self._id = self.connect("value-changed", self.__volume_changed, player)
        self._id2 = player.connect("notify::volume", self.__volume_notify)
        self._id3 = player.connect("notify::mute", self.__mute_notify)
        self._orig_icon_list = self.props.icons
        player.notify("volume")
        player.notify("mute")

        self.connect("event", self._on_button_event, player)

        replaygain_menu = VolumeMenu(player)
        self.connect("popup-menu", self.__popup, replaygain_menu)
        connect_obj(
            self,
            "button-press-event",
            self.__volume_button_press,
            replaygain_menu,
            player,
        )

    def __popup(self, widget, menu):
        time = Gtk.get_current_event_time()
        button = 3
        qltk.popup_menu_under_widget(menu, widget, button, time)
        return True

    def __volume_button_press(self, menu, event, player):
        if event.type != Gdk.EventType.BUTTON_PRESS:
            return False

        if event.button == Gdk.BUTTON_SECONDARY:
            qltk.popup_menu_at_widget(menu, self, event.button, event.time)
            return True
        if event.button == Gdk.BUTTON_MIDDLE:
            # toggle the muted state, if the backend doesn't support it
            # this action will just be ignored
            player.mute = not player.mute
            return True
        return None

    def __iadd__(self, v):
        self.set_value(self.get_value() + v)
        return self

    def __isub__(self, v):
        self.set_value(self.get_value() - v)
        return self

    def __volume_changed(self, button, volume, player):
        player.handler_block(self._id2)
        player.volume = volume
        player.handler_unblock(self._id2)

    def __volume_notify(self, player, prop):
        self.handler_block(self._id)
        self.set_value(player.volume)
        self.handler_unblock(self._id)

    def __mute_notify(self, player, prop):
        self._update_mute(player)

    def _update_mute(self, player):
        if player.mute:
            # remove all icons except the mute one to show a muted state
            # that is not affected by the volume slider
            self.props.icons = [self._orig_icon_list[0]]
        else:
            self.props.icons = self._orig_icon_list

    def _on_button_event(self, widget, event, player):
        # pulsesink doesn't emit volume changes when it's paused, but
        # fetching the value works. To prevent user volume changes based on a
        # false starting point update the slider on any action on the
        # volume button.
        self.handler_block(self._id)
        self.set_value(player.volume)
        self.handler_unblock(self._id)
        # same with mute
        self._update_mute(player)


class VolumeMenu(Gtk.Menu):
    __modes = (
        ("auto", _("Auto_matic"), None),
        ("track", _("_Track Mode"), ["track"]),
        ("album", _("_Album Mode"), ["album", "track"]),
    )

    def __init__(self, player):
        super().__init__()

        # ubuntu 12.04..
        if hasattr(player, "bind_property"):
            # Translators: player state, no action
            item = Gtk.CheckMenuItem(label=_("_Mute"), use_underline=True)
            player.bind_property(
                "mute", item, "active", GObject.BindingFlags.BIDIRECTIONAL
            )
            self.append(item)
            item.show()

        item = Gtk.MenuItem(label=_("_Replay Gain Mode"), use_underline=True)
        self.append(item)
        item.show()

        # Set replaygain mode as saved in configuration
        replaygain_mode = config.gettext("player", "replaygain_mode", "auto")
        self.__set_mode(player, replaygain_mode)

        rg = Gtk.Menu()
        rg.show()
        item.set_submenu(rg)
        item = None
        for mode, title, _profile in self.__modes:
            item = RadioMenuItem(group=item, label=title, use_underline=True)
            rg.append(item)
            item.connect("toggled", self.__changed, player, mode)
            if replaygain_mode == mode:
                item.set_active(True)
            item.show()

    def __set_mode(self, player, mode):
        selected_mode = next((m for m in self.__modes if m[0] == mode), None)
        if selected_mode is None:
            print_e(f"Invalid selected replaygain mode: {mode!r}")
            selected_mode = self.__modes[0]
            print_e(f"Falling back to replaygain mode: {selected_mode[0]!r}")

        player.replaygain_profiles[0] = selected_mode[2]
        player.reset_replaygain()

    def __changed(self, item, player, mode):
        if item.get_active():
            config.settext("player", "replaygain_mode", mode)
            self.__set_mode(player, mode)

    def popup(self, *args):
        gain = config.getboolean("player", "replaygain")
        for child in self.get_children():
            child.set_sensitive(gain)
        return super().popup(*args)


class PlayPauseButton(Gtk.Button):
    __gsignals__: GSignals = {
        "toggled": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        super().__init__(relief=Gtk.ReliefStyle.NONE)
        self._pause_image = SymbolicIconImage(
            "media-playback-pause", Gtk.IconSize.LARGE_TOOLBAR
        )
        self._play_image = SymbolicIconImage(
            "media-playback-start", Gtk.IconSize.LARGE_TOOLBAR
        )
        self._set_active(False)
        self.connect("clicked", self._on_clicked)

    def _on_clicked(self, *args):
        self.set_active(not self.get_active())

    def _set_active(self, is_active):
        if self.get_child():
            self.remove(self.get_child())

        if is_active:
            self.add(self._pause_image)
        else:
            self.add(self._play_image)
        self.get_child().show()

        self.emit("toggled")

    def set_active(self, is_active):
        if self.get_active() == is_active:
            return
        config.set("player", "is_playing", is_active)
        self._set_active(is_active)

    def get_active(self):
        return self.get_child() is self._pause_image


class PlayControls(Gtk.VBox):
    def __init__(self, player, library):
        super().__init__(spacing=3)

        upper = Gtk.Table(n_rows=1, n_columns=3, homogeneous=True)
        upper.set_row_spacings(3)
        upper.set_col_spacings(3)

        prev = Gtk.Button(relief=Gtk.ReliefStyle.NONE)
        prev.add(SymbolicIconImage("media-skip-backward", Gtk.IconSize.LARGE_TOOLBAR))
        upper.attach(prev, 0, 1, 0, 1)

        play = PlayPauseButton()
        upper.attach(play, 1, 2, 0, 1)

        next_ = Gtk.Button(relief=Gtk.ReliefStyle.NONE)
        next_.add(SymbolicIconImage("media-skip-forward", Gtk.IconSize.LARGE_TOOLBAR))
        upper.attach(next_, 2, 3, 0, 1)

        lower = Gtk.Table(n_rows=1, n_columns=3, homogeneous=True)
        lower.set_row_spacings(3)
        lower.set_col_spacings(3)

        self.volume = Volume(player)
        self.volume.set_relief(Gtk.ReliefStyle.NONE)
        lower.attach(self.volume, 0, 1, 0, 1)

        # XXX: Adwaita defines a different padding for GtkVolumeButton
        # We force it to 0 here, which works because the other (normal) buttons
        # in the grid set the width/height
        qltk.add_css(
            self.volume,
            """
            .button {
                padding: 0px;
            }
        """,
        )

        seekbutton = SeekButton(player, library)
        seekbutton.set_relief(Gtk.ReliefStyle.NONE)
        lower.attach(seekbutton, 1, 3, 0, 1)

        self.pack_start(upper, False, True, 0)
        self.pack_start(lower, False, True, 0)

        connect_obj(prev, "clicked", self.__previous, player)
        self._toggle_id = play.connect("toggled", self.__playpause, player)
        play.add_events(Gdk.EventMask.SCROLL_MASK)
        connect_obj(play, "scroll-event", self.__scroll, player)
        connect_obj(next_, "clicked", self.__next, player)
        connect_destroy(player, "song-started", self.__song_started, next_, play)
        connect_destroy(player, "paused", self.__on_set_paused_unpaused, play, False)
        connect_destroy(player, "unpaused", self.__on_set_paused_unpaused, play, True)

    def __on_set_paused_unpaused(self, player, button, state):
        # block to prevent a signal cycle in case the paused signal and state
        # get out of sync (shouldn't happen.. but)
        button.handler_block(self._toggle_id)
        button.set_active(state)
        button.handler_unblock(self._toggle_id)

    def __scroll(self, player, event):
        if event.direction in [Gdk.ScrollDirection.UP, Gdk.ScrollDirection.LEFT]:
            player.previous()
        elif event.direction in [Gdk.ScrollDirection.DOWN, Gdk.ScrollDirection.RIGHT]:
            player.next()

    def __song_started(self, player, song, next, play):
        play.set_active(not player.paused)

    def __playpause(self, button, player):
        if button.get_active():
            player.play()
        else:
            player.paused = True

    def __previous(self, player):
        player.previous()

    def __next(self, player):
        player.next()



# Copyright 2014 Christoph Reiter
#           2018 Ludovic Druette
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


class MMKeysImportError(ImportError):
    pass


class MMKeysAction:
    PLAY = "play"
    STOP = "stop"
    PAUSE = "pause"
    PREV = "prev"
    NEXT = "next"
    PLAYPAUSE = "playpause"
    FORWARD = "forward"
    REWIND = "rewind"
    REPEAT = "repeat"
    SHUFFLE = "shuffle"


class MMKeysBackend:
    def __init_(self, name, callback):
        """Callback will be called in the main thread and gets
        passed an MMKeysAction. `name` should be the application name.
        """

        raise NotImplementedError

    @classmethod
    def is_active(cls):
        """Should return if the backend should be used"""
        return True

    def grab(self):
        """Should tell the backend that the application was active
        (e.g. the main window got focused)
        """

    def cancel(self):
        """After cancel returns the callback will no longer be called.
        Can be called multiple times.
        """
        raise NotImplementedError



# Copyright 2022 Thomas Leberbauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from gi.repository import GObject, Gio, GdkPixbuf, Gtk, Pango, Gdk
from cairo import Surface
from .models import AlbumListItem

from quodlibet.qltk.cover import get_no_cover_pixbuf
from quodlibet.qltk.image import add_border_widget, get_surface_for_pixbuf
from quodlibet.util import DeferredSignal


def _no_cover(size, widget) -> Surface | None:
    old_size, surface = getattr(_no_cover, "cache", (None, None))
    if old_size != size or surface is None:
        surface = get_surface_for_pixbuf(widget, get_no_cover_pixbuf(size, size))
        _no_cover.cache = size, surface  # type: ignore
    return surface


class AlbumWidget(Gtk.FlowBoxChild):
    """An AlbumWidget displays an album with a cover and a label.

    The cover initially holds a placeholder. When the widget is drawn the real
    cover loads and the label is shown.
    """

    __gsignals__ = {"songs-menu": (GObject.SignalFlags.RUN_LAST, None, ())}

    padding = GObject.Property(type=int, default=0)
    cover_size = GObject.Property(type=int, default=48)
    text_visible = GObject.Property(type=bool, default=True)
    display_pattern = GObject.Property()

    def __init__(
        self, model: AlbumListItem, cancelable: Gio.Cancellable | None = None, **kwargs
    ):
        super().__init__(has_tooltip=True, **kwargs)

        self.model = model
        self._cancelable = cancelable
        self.__draw_handler_id = None

        self._box = box = Gtk.Box(vexpand=False, orientation=Gtk.Orientation.VERTICAL)

        image_size = self.__get_image_size()
        self._image = Gtk.Image(width_request=image_size, height_request=image_size)
        self._label = label = Gtk.Label(
            ellipsize=Pango.EllipsizeMode.END, justify=Gtk.Justification.CENTER
        )

        box.pack_start(self._image, True, True, 0)
        box.pack_start(self._label, True, True, 0)

        eb = Gtk.EventBox()
        eb.connect("popup-menu", lambda _: self.emit("songs-menu"))
        eb.connect("button-press-event", self.__rightclick)
        eb.add(box)

        self.add(eb)

        # show all before binding "visible" so the label will stay hidden if so
        # configured by the "text_visible" property.
        self.show_all()

        self.bind_property("padding", box, "margin", GObject.BindingFlags.SYNC_CREATE)
        self.bind_property("padding", box, "spacing", GObject.BindingFlags.SYNC_CREATE)
        self.bind_property(
            "text-visible", label, "visible", GObject.BindingFlags.SYNC_CREATE
        )

        model.connect("notify::album", lambda *a: self._populate())
        model.connect("notify::label", lambda *a: self._set_text(model.label))
        model.connect("notify::cover", lambda *a: self._set_cover(model.cover))

        self.connect("query-tooltip", self.__tooltip)
        self.connect("notify::cover-size", self.__cover_size)
        self.connect("notify::display-pattern", self.__display_pattern)

        self._set_cover(self.model.cover)
        self._set_text(self.model.label)
        self._populate_on_draw()

    def do_get_preferred_width(self):
        image_size = self.__get_image_size()
        width = image_size + 4 * self.props.padding
        return (width, width)

    def __get_image_size(self) -> int:
        return self.props.cover_size + 2

    def populate(self):
        self._populate_on_draw()

    def _populate_on_draw(self):
        self.__draw_handler_id = self._image.connect(
            "draw", DeferredSignal(self.__draw, timeout=10)
        )

    def __draw(self, widget, cr):
        if self.__draw_handler_id is None:
            return
        self._image.disconnect(self.__draw_handler_id)
        self.__draw_handler_id = None
        self._populate()

    def _populate(self):
        size = self.props.scale_factor * self.props.cover_size
        self.model.load_cover(size, self._cancelable)
        self.model.format_label(self.props.display_pattern)

    def _set_cover(self, cover: GdkPixbuf.Pixbuf | None = None):
        if cover:
            pb = add_border_widget(cover, self)
            surface = get_surface_for_pixbuf(self, pb)
        else:
            size = self.props.scale_factor * self.props.cover_size
            surface = _no_cover(size, self)
        self._image.props.surface = surface

    def _set_text(self, label: str | None = None):
        if label:
            self._label.set_markup(label)

    def __cover_size(self, _, prop):
        size = self.__get_image_size()
        self._image.props.width_request = size
        self._image.props.height_request = size
        self._set_cover()
        self._populate_on_draw()

    def __display_pattern(self, _, prop):
        self.model.format_label(self.props.display_pattern)

    def __rightclick(self, widget, event):
        if event.button == Gdk.BUTTON_SECONDARY:
            self.emit("songs-menu")

    def __tooltip(self, widget, x, y, keyboard_tip, tooltip):
        label = self.model.label
        if label:
            tooltip.set_markup(label)
        return True



# Copyright 2004-2008 Joe Wreschnig
#           2009-2024 Nick Boultbee
#           2011-2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Simple proxy to a Python ConfigParser

ConfigParser uses "str" on both Python 2/3, on Python 2 we simply encode
text/windows paths using utf-8 and save bytes as is.

On Python 3 we save text as is, convert paths to bytes/utf-8 and convert bytes
to unicode with utf-8/surrogateescape.

The final representation on disk should then in both cases be the same.
"""

import os
import csv
import collections
from io import StringIO

from configparser import RawConfigParser as ConfigParser, Error, NoSectionError

from quodlibet import print_w
from senf import fsnative

from quodlibet.util import list_unique, print_d
from quodlibet.util.atomic import atomic_save
from quodlibet.util.string import join_escape, split_escape
from quodlibet.util.path import mkdir


# In newer RawConfigParser it is possible to replace the internal dict. The
# implementation only uses items() for writing, so replace with a dict that
# returns them sorted. This makes it easier to look up entries in the file.
class _sorted_dict(collections.OrderedDict):  # noqa
    def items(self):
        return sorted(super().items())


_DEFAULT = object()


class Config:
    """A wrapper around RawConfigParser.

    Provides a ``defaults`` attribute of the same type which can be used
    to set default values.
    """

    def __init__(self, version=None, _defaults=True):
        """Use read() to read in an existing config file.

        version should be an int starting with 0 that gets incremented if you
        want to register a new upgrade function. If None, upgrade is disabled.
        """

        self._config = ConfigParser(dict_type=_sorted_dict)
        self.defaults = None
        if _defaults:
            self.defaults = Config(_defaults=False)
        self._version = version
        self._loaded_version = None
        self._upgrade_funcs = []

    def _do_upgrade(self, func):
        assert self._loaded_version is not None
        assert self._version is not None

        old_version = self._loaded_version
        new_version = self._version
        if old_version != new_version:
            print_d("Config upgrade: %d->%d (%r)" % (old_version, new_version, func))
            func(self, old_version, new_version)

    def get_version(self):
        """Get the version of the loaded config file (for testing only)

        Raises Error if no file was loaded or versioning is disabled.
        """

        if self._version is None:
            raise Error("Versioning disabled")

        if self._loaded_version is None:
            raise Error("No file loaded")

        return self._loaded_version

    def register_upgrade_function(self, function):
        """Register an upgrade function that gets called at each read()
        if the current config version and the loaded version don't match.

        Can also be registered after read was called.

        function(config, old_version: int, new_version: int) -> None
        """

        if self._version is None:
            raise Error("Versioning disabled")

        self._upgrade_funcs.append(function)
        # after read(), so upgrade now
        if self._loaded_version is not None:
            self._do_upgrade(function)
        return function

    def reset(self, section, option):
        """Reset the value to the default state"""

        assert self.defaults is not None

        try:
            self._config.remove_option(section, option)
        except NoSectionError:
            pass

    def options(self, section):
        """Returns a list of options available in the specified section."""

        try:
            options = self._config.options(section)
        except NoSectionError:
            if self.defaults:
                return self.defaults.options(section)
            raise
        else:
            if self.defaults:
                try:
                    options.extend(self.defaults.options(section))
                    options = list_unique(options)
                except NoSectionError:
                    pass
            return options

    def get(self, section, option, default=_DEFAULT):
        """get(section, option[, default]) -> str

        If default is not given or set, raises Error in case of an error
        """

        try:
            return self._config.get(section, option)
        except Error as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.get(section, option)
                    except Error:
                        pass
                raise
            if "No section:" in str(e):
                print_w(f"Config problem: {e}")
            return default

    def gettext(self, *args, **kwargs):
        value = self.get(*args, **kwargs)
        # make sure there are no surrogates
        value.encode("utf-8")
        return value

    def getbytes(self, section, option, default=_DEFAULT):
        try:
            value = self._config.get(section, option)
            return value.encode("utf-8", "surrogateescape")
        except (Error, ValueError) as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getbytes(section, option)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def getboolean(self, section, option, default=_DEFAULT):
        """getboolean(section, option[, default]) -> bool

        If default is not given or set, raises Error in case of an error
        """

        try:
            return self._config.getboolean(section, option)
        except (Error, ValueError) as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getboolean(section, option)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def getint(self, section, option, default=_DEFAULT):
        """getint(section, option[, default]) -> int

        If default is not give or set, raises Error in case of an error
        """

        try:
            return int(self._config.getfloat(section, option))
        except (Error, ValueError) as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getint(section, option)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def getfloat(self, section, option, default=_DEFAULT):
        """getfloat(section, option[, default]) -> float

        If default is not give or set, raises Error in case of an error
        """

        try:
            return self._config.getfloat(section, option)
        except (Error, ValueError) as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getfloat(section, option)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def getstringlist(self, section, option, default=_DEFAULT):
        """getstringlist(section, option[, default]) -> list

        If default is not given or set, raises Error in case of an error.
        Gets a list of strings, using CSV to parse and delimit.
        """

        try:
            value = self._config.get(section, option)

            parser = csv.reader([value], lineterminator="\n", quoting=csv.QUOTE_MINIMAL)
            try:
                vals = next(parser)
            except (csv.Error, ValueError) as e:
                raise Error(e) from e
            return vals
        except Error as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getstringlist(section, option)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def setstringlist(self, section, option, values):
        """Saves a list of unicode strings using the csv module"""

        sw = StringIO()
        values = [str(v) for v in values]

        writer = csv.writer(sw, lineterminator="\n", quoting=csv.QUOTE_MINIMAL)
        writer.writerow(values)
        self.set(section, option, sw.getvalue())

    def setlist(self, section, option, values, sep=","):
        """Saves a list of str using ',' as a separator and \\ for escaping"""

        values = [str(v) for v in values]
        joined = join_escape(values, sep)
        self.set(section, option, joined)

    def getlist(self, section, option, default=_DEFAULT, sep=","):
        """Returns a str list saved with setlist()"""

        try:
            value = self._config.get(section, option)
            return split_escape(value, sep)
        except (Error, ValueError) as e:
            if default is _DEFAULT:
                if self.defaults is not None:
                    try:
                        return self.defaults.getlist(section, option, sep=sep)
                    except Error:
                        pass
                raise Error(e) from e
            return default

    def set(self, section, option, value):
        """Saves the string representation for the passed value

        Don't pass unicode, encode first.
        """

        if isinstance(value, bytes):
            raise TypeError("use setbytes")

        # RawConfigParser only allows string values but doesn't
        # scream if they are not (and it only fails before the
        # first config save..)
        if not isinstance(value, str):
            value = str(value)

        try:
            self._config.set(section, option, value)
        except NoSectionError:
            if self.defaults and self.defaults.has_section(section):
                self._config.add_section(section)
                self._config.set(section, option, value)
            else:
                raise

    def settext(self, section, option, value):
        value = str(value)

        # make sure there are no surrogates
        value.encode("utf-8")

        self.set(section, option, value)

    def setbytes(self, section, option, value):
        assert isinstance(value, bytes)

        value = value.decode("utf-8", "surrogateescape")

        self.set(section, option, value)

    def write(self, filename):
        """Write config to filename.

        Can raise EnvironmentError
        """

        assert isinstance(filename, fsnative)

        mkdir(os.path.dirname(filename))

        # temporary set the new version for saving
        if self._version is not None:
            self.add_section("__config__")
            self.set("__config__", "version", self._version)
        try:
            with atomic_save(filename, "wb") as fileobj:
                temp = StringIO()
                self._config.write(temp)
                data = temp.getvalue().encode("utf-8", "surrogateescape")
                fileobj.write(data)
        finally:
            if self._loaded_version is not None:
                self.set("__config__", "version", self._loaded_version)

    def clear(self):
        """Remove all sections."""

        for section in self._config.sections():
            self._config.remove_section(section)

    def is_empty(self):
        """Whether the config has any sections"""

        return not self._config.sections()

    def read(self, filename):
        """Reads the config from `filename` if the file exists,
        otherwise does nothing

        Can raise EnvironmentError, Error.
        """

        try:
            with open(filename, "rb") as fileobj:
                fileobj = StringIO(fileobj.read().decode("utf-8", "surrogateescape"))
                self._config.read_file(fileobj, filename)
        except OSError:
            print_d(f"No config file found at {filename} – using defaults")
            return

        # don't upgrade if we just created a new config
        if self._version is not None:
            self._loaded_version = self.getint("__config__", "version", -1)
            for func in self._upgrade_funcs:
                self._do_upgrade(func)

    def has_option(self, section, option):
        """If the given section exists, and contains the given option"""

        return self._config.has_option(section, option) or (
            self.defaults and self.defaults.has_option(section, option)
        )

    def has_section(self, section):
        """If the given section exists"""

        return self._config.has_section(section) or (
            self.defaults and self.defaults.has_section(section)
        )

    def remove_option(self, section, option):
        """Remove the specified option from the specified section

        Can raise Error.
        """

        return self._config.remove_option(section, option)

    def add_section(self, section):
        """Add a section named section to the instance if it not already
        exists."""

        if not self._config.has_section(section):
            self._config.add_section(section)


class ConfigProxy:
    """Provides a Config object with a fixed section and a possibility to
    prefix option names in that section.

    e.g. it can create a view of the "plugin" section and prefix all
    options with a plugin name.
    """

    def __init__(self, real_config, section_name, _defaults=True):
        self._real_config = real_config
        self._section_name = section_name
        self.defaults = None
        if _defaults:
            self.defaults = self._new_defaults(real_config.defaults)

    def _new_defaults(self, real_default_config):
        return ConfigProxy(real_default_config, self._section_name, False)

    def _option(self, name):
        """Override if you want to change option names. e.g. prefix them"""

        return name

    @classmethod
    def _init_wrappers(cls):
        def get_func(name):
            def method(self, option, *args, **kwargs):
                config_getter = getattr(self._real_config, name)
                return config_getter(
                    self._section_name, self._option(option), *args, **kwargs
                )

            return method

        # methods starting with a section arg
        for name in [
            "get",
            "set",
            "getboolean",
            "getint",
            "getfloat",
            "reset",
            "settext",
            "gettext",
            "getbytes",
            "setbytes",
        ]:
            setattr(cls, name, get_func(name))


ConfigProxy._init_wrappers()



# Copyright (C) 2018 Eoin O'Neill (eoinoneill1991@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk
from quodlibet import _, util
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Icons, ToggleButton
from quodlibet.plugins.gui import UserInterfacePlugin


class RatingBox(Gtk.VBox):
    def __init__(self):
        super().__init__(self)

        self.thumb_ups = 1
        self.thumb_downs = 1

        self.title = Gtk.Label("")
        self.title.set_line_wrap(True)
        self.title.set_lines(2)

        hbox = Gtk.HBox()
        self.upvote = ToggleButton("👍")
        self.downvote = ToggleButton("👎")
        self.upvote.connect("toggled", self.__thumb_toggled)
        self.downvote.connect("toggled", self.__thumb_toggled)
        self.score_label = Gtk.Label("----")
        self.upvote.set_property("height-request", 50)
        self.downvote.set_property("height-request", 50)
        hbox.pack_start(self.upvote, True, True, 5)
        hbox.pack_start(self.downvote, True, True, 5)

        self.hbox = hbox
        self.pack_start(self.title, False, False, 10)
        self.pack_start(self.score_label, True, True, 5)
        self.pack_start(self.hbox, False, False, 5)

    def set_current_title(self, title):
        self.title.set_text(title)

    def set_current_score(self, cth_up, cth_down):
        self.thumb_ups = cth_up
        self.thumb_downs = cth_down
        self.__set_pending_score_value(self.thumb_ups - self.thumb_downs)

    def poll_vote(self, reset=True):
        upward = 1 if self.upvote.get_active() else 0
        downward = 1 if self.downvote.get_active() else 0
        vote = (upward, downward)
        if reset:
            self.downvote.set_active(False)
            self.upvote.set_active(False)
        return vote

    def __set_pending_score_value(self, score):
        existing_score = self.thumb_ups - self.thumb_downs
        if score == existing_score:
            self.score_label.set_markup(util.bold(str(int(score))))
        elif score > existing_score:
            self.score_label.set_markup(
                '<b><span foreground="green">' + str(int(score)) + "</span></b>"
            )
        else:
            self.score_label.set_markup(
                '<b><span foreground="red">' + str(int(score)) + "</span></b>"
            )

    def __thumb_toggled(self, button):
        if button.get_active():
            if button == self.upvote:
                self.downvote.set_active(False)
            elif button == self.downvote:
                self.upvote.set_active(False)

        vote = self.poll_vote(False)
        self.__set_pending_score_value(
            self.thumb_ups + vote[0] - self.thumb_downs - vote[1]
        )


class ThumbRating(EventPlugin, UserInterfacePlugin):
    """Plugin for more hands off rating system using a
    thumb up / thumbdown system."""

    PLUGIN_ID = "Thumb Rating"
    PLUGIN_NAME = _("Thumb Rating")
    PLUGIN_DESC_MARKUP = _(
        "Adds a thumb-up / thumb-down scoring system "
        "which is converted to a rating value. Useful "
        "for keeping running vote totals and sorting by "
        "<b><tt>~#score</tt></b>."
    )
    PLUGIN_ICON = Icons.USER_BOOKMARKS

    # Threshold value where points should be recalculated
    score_point_threshold = 0.2

    def enabled(self):
        self.rating_box = RatingBox()

    def create_sidebar(self):
        vbox = Gtk.VBox()
        vbox.pack_start(self.rating_box, False, False, 0)
        vbox.show_all()
        return vbox

    def disabled(self):
        self.rating_box.destroy()

    def plugin_on_song_ended(self, song, stopped):
        if song is not None:
            poll = self.rating_box.poll_vote()
            if poll[0] >= 1 or poll[1] >= 1:
                ups = int(song.get("~#wins") or 0)
                downs = int(song.get("~#losses") or 0)
                ups += poll[0]
                downs += poll[1]
                song["~#wins"] = ups
                song["~#losses"] = downs
                song["~#rating"] = ups / max((ups + downs), 2)
                # note: ^^^ Look into implementing w/ confidence intervals!
                song["~#score"] = ups - downs

    def plugin_on_song_started(self, song):
        if song is not None:
            ups = int(song("~#wins") or 0)
            downs = int(song("~#losses") or 0)

            # Handle case where there's no score but user has a defined rating.
            if (ups + downs == 0) and (song.get("~#rating")):
                percent = song["~#rating"]
                ups = int(percent * 10)
                downs = int((1.0 - percent) * 10.0)
                song["~#wins"] = ups
                song["~#losses"] = downs
            elif (
                song.get("~#rating")
                and abs((ups / max((ups + downs), 2)) - song["~#rating"])
                > self.score_point_threshold
            ):
                # Cases where rating and points are not in alignment.
                total = max(ups + downs, 10)
                percent = song["~#rating"]
                ups = int(percent * total)
                downs = int((1.0 - percent) * total)
                song["~#wins"] = ups
                song["~#losses"] = downs

            self.rating_box.set_current_score(ups, downs)
            self.rating_box.set_current_title(song("~artist~title"))
